"""splinepy/helpme/ffd.py.

Freeform Deformation!
"""

import gustaf as _gus

from splinepy import settings as _settings
from splinepy._base import SplinepyBase as _SplinepyBase
from splinepy.helpme.create import from_bounds as _from_bounds
from splinepy.spline import Spline as _Spline


def _rescaled_vertices(vertices, scale, offset):
    return (vertices * (1 / scale)) + offset


def _spline_para_dim_and_mesh_dim_matches(spline, mesh):
    return spline.para_dim == mesh.vertices.shape[1]


class FFD(_SplinepyBase):
    """
    Free-form deformation is a method used to deform an object by a
    deformation function. In our case the object is given via a mesh, the
    currently supported mesh-types are gustaf.Vertices and its subclasses,
    and the deformations function by a spline, supported splines are
    splinepy.Spline and its subclasses.
    The splines parametric dimension will be scaled in to a unit-hypercube
    as well as the original meshes vertices.

    The FFD class provides functions to modify the spline by completely
    overwriting the spline whole spline or parts of it. To obtain the
    deformed mesh mapped into the latest spline, retrieve the mesh
    attribute.

    Parameters
    ----------
    mesh: gustaf.Vertices
      Mesh used in the FFD. Defaults to None.
    spline: splinepy.Spline
      Spline used in the FFD. Defaults to None.
    padding: float
      Padding factor to scale mesh into spline's parametric space.
      Ideally, this should be 0, but to avoid floating point issues during
      evaluation, we apply default settings.TOLERANCE

    Returns
    -------
    None
    """

    def __init__(
        self,
        mesh=None,
        spline=None,
        padding=None,
    ):
        # init variables
        self._mesh = None
        self._spline = None
        self._q_vertices = None
        self._q_scale = None
        self._q_offset = None

        # use setters for attr
        if padding is None:
            self.padding = _settings.TOLERANCE
        if mesh is not None:
            self.mesh = mesh
        if spline is not None:
            self.spline = spline

    @property
    def mesh(self):
        """Returns copy of current mesh. Before copying, it applies
        deformation.

        Returns
        -------
        current_mesh: gustaf.Vertices or derived
            Current Mesh with the deformation according to the current spline.
        """
        # is there mesh?
        if self._mesh is None:
            raise ValueError("Please set mesh first.")

        # evaluate new vertices
        current_mesh = type(self._mesh)(
            vertices=self._spline.evaluate(self._q_vertices)
        )

        # apply connectivity if applicable
        if hasattr(self._mesh, "elements"):
            current_mesh.elements = self._mesh.elements.copy()

        return current_mesh

    @mesh.setter
    def mesh(self, mesh):
        """Sets mesh. If it is first time, the copy of it will be saved as
        original mesh. If spline is already defined and in transformed status,
        it applies transformation directly.

        Parameters
        -----------
        mesh: gustaf.Vertices
            Mesh used for the FFD

        Returns
        --------
        None
        """
        if mesh is None:
            self._mesh = None
            self._q_vertices = None
            self._q_offset = None
            self._q_scale = None
            return None

        # type check
        if not isinstance(mesh, _gus.Vertices):
            raise TypeError("mesh should be an instance of gus.Vertices")

        # warn and erase existing spline if dimension doesn't match
        if (
            self._spline is not None
            and not _spline_para_dim_and_mesh_dim_matches(self._spline, mesh)
        ):
            self._logw(
                "Mismatch between mesh vertices' dim and set spline's",
                "para_dim.",
                "Deleting existing spline",
            )

            self._spline = None

        # create a spline if there isn't one
        if self._spline is None:
            # Define a default spline if mesh is given but no spline
            par_dim = mesh.vertices.shape[1]
            # try to expand bounds same ratio as
            bounds = mesh.bounds().copy()
            self.spline = _from_bounds([[0] * par_dim, [1] * par_dim], bounds)
            self._logd("created default spline")

        self._logd("Setting mesh.")
        self._logd("Mesh Info:")
        self._logd(f"  Vertices: {mesh.vertices.shape}.")
        self._logd(f"  Bounds: {mesh.bounds()}.")

        # keep original copy
        self._mesh = mesh.copy()

        self._logd("Fitting mesh into spline's parametric space.")

        self._q_vertices = self._mesh.vertices.copy()

        # get mesh bound to scale
        scaling_bounds = self._mesh.bounds().copy()

        # apply padding
        scaling_bounds[0] -= self.padding
        scaling_bounds[1] += self.padding

        # save mesh offset and scale
        self._q_offset = scaling_bounds[0]
        self._q_scale = 1 / (scaling_bounds[1] - scaling_bounds[0])

        # scale and offset vertices coordinates
        self._q_vertices -= self._q_offset
        self._q_vertices *= self._q_scale

        self._logd("Successfully scaled mesh vertices!")

    @property
    def spline(self):
        """Returns a copy of the spline. Please use the setter to explicitly
        make changes to the spline.

        Parameters
        -----------
        None

        Returns
        --------
        self._spline: Spline
        """
        return self._spline

    @spline.setter
    def spline(self, spline):
        """Sets spline. The spline parametric range bounds will be converted
        into the bounds [0,1]^para_dim.

        Parameters
        -----------
        spline: Spline
            New Spline for the next deformation

        Returns
        --------
        None
        """
        if spline is None:
            self._spline = None
            return None

        if not isinstance(spline, _Spline):
            raise TypeError(
                "spline should be an instance of splinepy.Spline class"
            )

        if (
            self._mesh is not None
            and not _spline_para_dim_and_mesh_dim_matches(spline, self._mesh)
        ):
            raise ValueError(
                "Spline's para_dim should match mesh vertices' dim."
            )

        # copy input
        self._spline = spline.copy()

        # normalize if needed
        if spline.has_knot_vectors:
            self._spline.normalize_knot_vectors()

    @property
    def padding(self):
        """
        Padding ratio for fitting mesh inside spline's parametric dimension.
        It should be a positive number.

        Parameters
        ----------
        None

        Returns
        -------
        padding: float
        """
        return self._padding

    @padding.setter
    def padding(self, padding):
        """
        Padding Setter.
        """
        if padding < 0:
            raise ValueError("Padding should be a positive number.")

        if padding < _settings.TOLERANCE:
            raise ValueError(
                "Padding is too small - "
                "should be bigger than the tolerance ({settings.TOLERANCE})."
            )

        self._padding = padding

        # if there's q_vertices, reset mesh, so that new q_vertices are formed
        if self._q_vertices is not None:
            self.mesh = self._mesh

    def show(self, **kwargs):
        """Visualize. Shows the deformed mesh and the current spline. Currently
        visualization is limited to vedo.

        Parameters
        ----------
        title: str
            Title of the vedo window. Defaults to "gustaf - FFD".
        return_showable: bool
            If true returns a dict of the showable items. Defaults to False.
        return_gustaf: bool
            Return dict of gustaf discrete objects, for example,
            {Vertices, Edges, Faces}, instead of opening a window.
            Defaults to False.
        kwargs: Any
            Arbitrary keyword arguments. These are passed onto the vedo
            functions. Please be aware, that no checking of these are performed
            in this function.

        Returns
        -------
        Any:
            Returns, if applicable, the vedo plotter. 'close=False' as argument
            to get the plotter.
        """
        if self._spline is None and self._mesh is None:
            raise ValueError("Please set a mesh before calling show()")
        return_showable = kwargs.pop("return_showable", False)
        return_gustaf = kwargs.pop("return_gustaf", False)
        title = kwargs.pop("title", "FFD")

        if return_gustaf and return_showable:
            raise ValueError(
                "Either one of following params can be True: "
                "{return_gustaf, return_showable} "
                "You've set both True."
            )

        things_to_show = self.showable(**kwargs)

        if return_gustaf or return_showable:
            return things_to_show

        return _gus.show(
            ["Original Mesh", things_to_show["original_mesh"]],
            [
                "Spline and Mesh",
                *things_to_show["spline"].values(),
                things_to_show["mesh"],
            ],
            ["Deformed Mesh", things_to_show["mesh"]],
            title=title,
            **kwargs,
        )

    def showable(self, **kwargs):
        """Returns a dictionary of showable items to describe the FFD at the
        current state.

        See show() for more information. This function redirects to it
        directly with the return_showable keyword set to True.
        """
        if self._spline is None and self._mesh is None:
            raise ValueError("Please set a mesh before calling show()")

        return_gustaf = kwargs.get("return_gustaf", False)

        things_to_show = {}

        # let's show faces at most, since volumes can take awhile
        mesh = self._mesh
        if mesh.kind == "volume":
            mesh = _gus.Faces(
                mesh.const_vertices,
                mesh.faces()[mesh.single_faces()],
            )

        # original mesh
        things_to_show["original_mesh"] = mesh

        # current spline
        self._spline.show_options["alpha"] = 0.3
        things_to_show["spline"] = self._spline

        # deformed mesh
        deformed = self.mesh
        things_to_show["mesh"] = deformed

        if return_gustaf:
            return things_to_show

        # let's turn everything into showable and return
        for k, v in things_to_show.items():
            things_to_show[k] = v.showable(**kwargs)

        return things_to_show



import numpy as _np

from splinepy import settings as _settings
from splinepy import spline as _spline
from splinepy import splinepy_core as _splinepy_core
from splinepy import utils as _utils


class BezierBase(_spline.Spline):
    r"""Bezier Base. Contains extra operations that are only
    available for bezier families.

    Passes all arguments to :code:`super.__init__()`, see :class:`.Spline`.

    Beziers are a special type of spline where the basis functions have
    a global support. The basis functions are given by the so-called
    *Berstein polynomials*

    .. math::
            B^{i;p}(u) = \binom{p}{i} u^i (1-u)^{p-i}

    The parametric domain of a Bezier spline is always a multi-dimensional
    hypercube, that is :math:`\Omega_{param}=[0,1]^{N_{param}}`.
    We refer to the documentation of the classes :class:`.Bezier` and
    :class:`.RationalBezier` for explanations how to construct splines using
    Berstein polynomials as basis.

    Moreover, we like to make the point that Bezier splines can also be seen
    as a special type of B-Splines with open knot vectors (i.e., the first and
    last entry are repeated :math:`p+1`-times) that do not feature additional
    internal knots.

    For usage examples, we once again refer to the derived classes
    :class:`.Bezier` and :class:`.RationalBezier`.
    """

    __slots__ = ()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __mul__(self, factor):
        """
        Overloads multiplication between splines with different types of
        degrees

        The resulting spline fulfills the equation
          factora(t) * factorb(t) = result(t)

        Parameters
        ----------
        factor :  float or BezierBase
          Spline with compatible dimensionality

        Returns
        -------
        multiplied : Bezier or RationalBezier
          New spline with updated types
        """
        # scalar
        if isinstance(factor, float):
            return type(self)(
                control_points=self.control_points * factor,
                degrees=self.degrees,
            )

        # Supports only Bezier families
        if not isinstance(factor, BezierBase):
            raise TypeError(
                f"Multiplication with {type(factor)}-type is not supported."
            )

        # multiply - dimension compatibility is checked in cpp side
        multiplied = _splinepy_core.multiply(self, factor)

        # return corresponding type
        return _settings.NAME_TO_TYPE[multiplied.name](spline=multiplied)

    def __add__(self, summand):
        """
        Calculates the spline that forms the sum of the summand and the
        current spline (function argument)

        The resulting spline fulfils the equation
          self(t) + summand(t) = result(t)

        Parameters
        ----------
        summand: type(self)
          spline with same parametric and physical dimension

        Returns
        -------
        added: type(self)
          New spline that describes sum
        """
        # same type check, same para dim check, same dim check done in cpp
        added = _splinepy_core.add(self, summand)

        return type(self)(spline=added)

    def derivative_spline(self, orders):
        r"""
        Calculates the spline that describes the requested derivative in a
        close-form representation of the the same spline-type.

        Parameters
        ----------
        orders: (para_dim,) array-like
          requested orders

        Returns
        -------
        derivative_spline : BezierBase
        """
        # Ensure array-type
        orders = _utils.data.enforce_contiguous(orders, dtype="int32")
        # dimension compatibility checked in cpp
        derivative_spline = _splinepy_core.derivative_spline(self, orders)

        return _settings.NAME_TO_TYPE[derivative_spline.name](
            spline=derivative_spline
        )

    def compose(self, inner_function, compute_sensitivities=False):
        r"""
        Calculates the spline that forms the composition of the inner function
        spline (function argument), using the caller spline as the outer (or
        deformation function).

        Given an outer function :math:`\mathcal{T}` and an inner function
        (spline) :math:`\mathcal{K}`, this function returns a new spline
        function :math:`\mathcal{M}` that represents the composition:

        .. math:: \mathcal{M}(\mathbf{u}) =
                  \mathcal{T}\big(\mathcal{K}(\mathbf{u})\big)
                  \quad \forall \mathbf{u}

        If the flag `compute_sensitivities` is set, compose also returns the
        derivative of the composed spline with respect to the outer function's
        control point positions. This corresponds to the basis function
        representation (which is given by a scalar valued spline).
        This functionality differs from :func:`.composition_derivative`, which
        computes the derivatives concerning the inner functions geometric
        parametrization.

        Given an outer function :math:`\mathcal{T}` with control points
        :math:`C_i` and an inner function :math:`\mathcal{K}`. This function
        computes:

        .. math:: \frac{\partial\mathcal{T}(\mathcal{K})}{\partial C_i}
                  (\mathbf{u}) =
                  B_i\big(\mathcal{K}\big)(\mathbf{u}) \quad \forall \mathbf{u}

        where :math:`B_i` represents the basis function associated to control
        point :math:`C_i`.


        Parameters
        ----------
        inner_function : BezierBase
          Bezier-Type Spline that represents inner function
        compute_sensitivities : bool
          Flag for return values. If set to true returns also sensitivities,
          default False

        Returns
        -------
        composed : BezierBase
          Composed function
        sensitivities : list (optional)
          List all splines that represents the derivatives with respect to
          internal control points
        """
        # dimension compatibility checked in cpp
        composed = _splinepy_core.compose(self, inner_function)

        if compute_sensitivities:
            composed_sensivities = _splinepy_core.compose_sensitivities(
                self, inner_function
            )

            return (
                _settings.NAME_TO_TYPE[composed.name](spline=composed),
                [
                    _settings.NAME_TO_TYPE[cc.name](spline=cc)
                    for cc in composed_sensivities
                ],
            )
        else:
            return _settings.NAME_TO_TYPE[composed.name](spline=composed)

    def composition_derivative(self, inner, inner_derivative):
        r"""
        Derivative of composition when given the differentiated inner function
        with constant outer function. This function differs from
        :func:`.compose_sensitivities`, which computes the derivatives
        concerning the outer function's control point positions.

        Given an outer function :math:`\mathcal{T}` and an inner function
        :math:`\mathcal{K}` with its derivative with respect to some design
        variable :math:`\alpha`, i.e.,
        :math:`\frac{\partial \mathcal{K}}{\partial \alpha}`, this function
        returns the derivative of the composed geometry in the form
        :math:`\frac{\partial \mathcal{T}(\mathcal{K})}{\partial\alpha}` by
        computing.

        .. math:: \frac{\partial\mathcal{T}(\mathcal{K})}{\partial\alpha} =
                  \frac{\partial \mathcal{T}}{\partial u}(\mathcal{K})\cdot
                  \frac{\partial \mathcal{K}}{\partial \alpha}

        Here :math:`u` refers to the parametric coordinates of the deformation
        function.

        Parameters
        ----------
        inner: BezierBase
        inner_derivative: BezierBase

        Returns
        -------
        composition_der: BezierBase
        """
        # compatibility check is done in cpp
        composition_der = _splinepy_core.composition_derivative(
            self, inner, inner_derivative
        )

        return _settings.NAME_TO_TYPE[composition_der.name](
            spline=composition_der
        )

    def split(self, para_dim, locations):
        """
        Splits spline at given locations along the given para_dim.

        Parameters
        ----------
        para_dim: int
        locations: array-like
          Should be in range of (0, 1)

        Returns
        -------
        split: list
          list of split splines. Self stays intact.
        """
        if max(locations) > 1 or min(locations) < 0:
            raise ValueError("Invalid split location. Should be in (0, 1).")
        split = _splinepy_core.split(self, para_dim, locations)

        return [type(self)(spline=s) for s in split]

    def extract_dim(self, dim):
        """
        Extracts a single physical dimension of a spline.

        Parameters
        ----------
        dim: int

        Returns
        -------
        extracted: type(self)
        """
        if dim >= self.dim:
            raise ValueError(f"Can't extract ({dim}) dim from {self.whatami}.")

        return type(self)(spline=_splinepy_core.extract_dim(self, dim))


class Bezier(BezierBase):
    r"""
    Bezier (Spline).

    Passes all arguments to :code:`super.__init__()`, see :class:`.BezierBase`.

    We can distinguish between different types of splines depending on the
    dimension of the parametric space.

    .. note::
        For simplicity, we restrict ourselves to the three most common types
        of splines in the following, namely curves, surfaces and volumes,
        although :code:`splinepy` also supports higher dimensions, see
        the documentation of :class:`.Spline` for more information.

    1. A spline of degree :math:`p` with :math:`(p+1)` control points
    :math:`P^i\in\mathbb{R}^{N_{phys}}` and a one-dimensional parameter space
    (i.e., :math:`N_{param}=1`) corresponds to a line embedded into the
    physical space:

    .. math::
            C(u) = \sum_{i=0}^{l} B^{i;p}(u) P^i

    2. A spline of degrees :math:`p,q` with :math:`(p+1)\cdot(q+1)` control
    points :math:`P^{i,j}\in\mathbb{R}^{N_{phys}}` and a two-dimensional
    parameter space (i.e., :math:`N_{param}=2`) corresponds to a surface,
    embedded into the physical space:

    .. math::
            S(u,v) = \sum_{i=0}^{l} \sum_{j=0}^{m} B^{i;p}(u) B^{j;q}(v)
                P^{i,j}

    Due to the tensor-product nature of the Bezier basis functions, this is
    often rewritten as in terms of multi-variate basis functions

    .. math::
            \tilde{B}^{i,j;p,q}(u,v) := B^{i;p}(u) B^{j;q}(v)

    3. A spline of degrees :math:`p,q,r` with :math:`(p+1)\cdot(q+1)\cdot(r+1)`
    control points :math:`P^{i,j,k}\in\mathbb{R}^{N_{phys}}` and a
    three-dimensional parameter space (i.e., :math:`N_{param}=3`) corresponds
    to a volume, embedded into the physical space:

    .. math::
            V^B(u,v,w) = \sum_{i=0}^{l} \sum_{j=0}^{m} \sum_{k=0}^{n}
                B^{i;p}(u) B^{j;q}(v) B^{k;r}(w) P^{i,j,k}

    Here, we can introduce the multi-variate basis functions

    .. math::
            \tilde{B}^{i,j,k;p,q,r}(u,v,w) := B^{i;p}(u) B^{j;q}(v)
                B^{k;r}(w)

    Higher-dimensional instances are constructed accordingly.

    **Usage**:

    .. code-block:: python

        # Polynomial Bezier surface
        polynomial_bezier = splinepy.Bezier(
            degrees=[2, 1],
            control_points=[
                [0.0, 0.0],
                [1.0, 0.0],
                [2.0, 1.0],
                [0.0, 2.0],
                [1.0, 1.0],
                [2.0, 2.0],
            ],
        )

    Parameters
    -----------
    degrees: (para_dim,) list-like
    control_points: (m, dim) list-like

    Returns
    --------
    None
    """

    __slots__ = ()

    def __init__(self, degrees=None, control_points=None, spline=None):
        super().__init__(
            spline=spline,
            degrees=degrees,
            control_points=control_points,
        )

    @property
    def bezier(self):
        return self.copy(saved_data=False)

    @property
    def rationalbezier(self):
        return _settings.NAME_TO_TYPE["RationalBezier"](
            **self.todict(), weights=_np.ones((self.cps.shape[0], 1))
        )

    @property
    def bspline(self):
        return _settings.NAME_TO_TYPE["BSpline"](
            spline=_splinepy_core.same_spline_with_knot_vectors(self)
        )

    @property
    def nurbs(self):
        return _settings.NAME_TO_TYPE["NURBS"](
            spline=_splinepy_core.same_spline_with_knot_vectors(
                self.rationalbezier
            )
        )



"""Supports single nurbs export for mfem.

Currently hardcoded for 2D-single-patch-splines.
"""

import gustaf as _gus
import numpy as _np

from splinepy.io.ioutils import dict_to_spline as _dict_to_spline
from splinepy.io.ioutils import form_lines as _form_lines
from splinepy.io.ioutils import make_meaningful as _make_meaningful
from splinepy.io.ioutils import next_line as _next_line

# keywords : possible assert value
_mfem_meaningful_keywords = {
    "MFEM NURBS mesh v1.0": "intro",
    "dimension": "para_dim",
    "elements": 1,
    "boundary": 4,
    "edges": 4,
    "vertices": 4,
    "knotvectors": 2,
    "weights": None,
    "FiniteElementSpace": None,
    "FiniteElementCollection": None,
    "VDim": "dim",
    "Ordering": 1,
}


def load(fname):
    """Reads mfem spline and returns a spline.

    Again, only supports 2D single patch.

    Parameters
    -----------
    fname: str

    Returns
    --------
    nurbs: dict
      dict ready to be used for init.
    """
    from copy import deepcopy as _deepcopy

    mk = _deepcopy(_mfem_meaningful_keywords)
    # they follow a strict order or keywords, so just gather those in order
    # Ordering is a hotkey because control points comes right after
    hotkeys = [
        "knotvectors",
        "weights",
        "Ordering",
    ]
    nurbs_dict = {"knotvectors": [], "weights": [], "Ordering": []}
    collect = False
    with open(fname) as f:
        for single_line in f:
            line = _make_meaningful(single_line)
            if not line:
                continue

            keyword_hit = [k for k in mk if line.startswith(k)]
            if len(keyword_hit) > 1:
                raise ValueError("double keyword hit!")

            elif len(keyword_hit) == 1 and keyword_hit[0] in hotkeys:
                collect = True
                current_key = _deepcopy(keyword_hit[0])
                continue

            elif len(keyword_hit) == 1 and keyword_hit[0] not in hotkeys:
                collect = False
                current_key = None

            elif len(keyword_hit) == 0 and collect:
                pass

            elif len(keyword_hit) == 0 and not collect:
                continue

            if collect:
                nurbs_dict[current_key].append(line)
            else:
                continue

    # parse nurbs_dict
    # kvs
    knot_vectors = nurbs_dict["knotvectors"][1:]
    knot_vectors[0] = eval("[" + knot_vectors[0].replace(" ", ",") + "]")
    knot_vectors[1] = eval("[" + knot_vectors[1].replace(" ", ",") + "]")
    # pop some values
    # ds
    degrees = [knot_vectors[0].pop(0), knot_vectors[1].pop(0)]
    ncps = knot_vectors[0].pop(0) * knot_vectors[1].pop(0)
    # ws
    weights = nurbs_dict["weights"]
    weights = [eval(w) for w in weights]  # hopefully not too slow
    # cps
    control_points = nurbs_dict["Ordering"]
    control_points = [
        eval(f"[{cp.replace(' ', ',')}]") for cp in control_points
    ]

    # double check
    # maybe separate them
    if (
        ncps != len(control_points)
        or ncps != len(weights)
        or int(nurbs_dict["knotvectors"][0]) != 2
    ):
        raise ValueError("Inconsistent spline info in " + fname)

    mfem_dict_spline = {
        "degrees": degrees,
        "knot_vectors": knot_vectors,
        "control_points": _np.ascontiguousarray(control_points),
        "weights": _np.ascontiguousarray(weights),
    }

    spline = _dict_to_spline([mfem_dict_spline])[0]

    _, reorder = dof_mapping(spline)

    spline.cps[:] = spline.cps[reorder]
    spline.ws[:] = spline.ws[reorder]

    return spline


def read_solution(
    fname,
    reference_nurbs,
):
    """
    Given solution and reference_nurbs, returns solution-nurbs-dict.

    Parameters
    -----------
    fname: str
    reference_nurbs: NURBS

    Returns
    --------
    solution_nurbs: dict
    """

    with open(fname) as f:
        hotkey = "Ordering"
        line = _next_line(f)
        solution = []
        collect = False
        while line is not None:
            if collect:  # check this first. should be true most time
                solution.append(eval(f"[{line.replace(' ', ',')}]"))
            elif hotkey in line:
                collect = True

            line = _next_line(f)

    if len(reference_nurbs.control_points) != len(solution):
        raise ValueError("Solution length does not match reference nurbs")

    mfem_dict_spline = reference_nurbs.todict()

    spline = _dict_to_spline([mfem_dict_spline])[0]

    _, reorder = dof_mapping(spline)

    spline.cps = _np.ascontiguousarray(solution)[reorder]
    spline.ws[:] = spline.ws[reorder]

    return spline


def _mfem_dof_map_2d(spline):
    """
    Dof mapping for 2D splines
    """
    mi = spline.multi_index

    to_m = []

    # corner vertices
    x_ids = [0, -1, -1, 0]
    y_ids = [0, 0, -1, -1]
    vertices = mi[x_ids, y_ids].tolist()

    # edge vertices
    edges = []
    edges.extend(mi[1:-1:, 0].tolist())  # bottom
    edges.extend(mi[1:-1, -1].tolist()[::-1])  # top
    edges.extend(mi[0, 1:-1].tolist())  # left
    edges.extend(mi[-1, 1:-1].tolist())  # right

    # internal
    internal = mi[1:-1, 1:-1].tolist()

    to_m.extend(vertices)
    to_m.extend(edges)
    to_m.extend(internal)

    return to_m, _np.argsort(to_m).tolist()


def _mfem_dof_map_3d(spline):
    """
    Dof mapping for 3D splines
    """
    mi = spline.multi_index

    to_m = []

    # corner vertices
    x_ids = [0, -1, -1, 0, 0, -1, -1, 0]
    y_ids = [0, 0, -1, -1, 0, 0, -1, -1]
    z_ids = [0, 0, 0, 0, -1, -1, -1, -1]

    vertices = mi[x_ids, y_ids, z_ids].tolist()

    # edges
    edges = []
    edges.extend(mi[1:-1, 0, 0].tolist())  # 1
    edges.extend(mi[1:-1, -1, 0].tolist()[::-1])  # 2
    edges.extend(mi[1:-1, 0, -1].tolist())  # 3
    edges.extend(mi[1:-1, -1, -1].tolist()[::-1])  # 4

    edges.extend(mi[0, 1:-1, 0].tolist())  # 5
    edges.extend(mi[-1, 1:-1, 0].tolist())  # 6
    edges.extend(mi[0, 1:-1, -1].tolist())  # 7
    edges.extend(mi[-1, 1:-1, -1].tolist())  # 8

    edges.extend(mi[0, 0, 1:-1].tolist())  # 9
    edges.extend(mi[-1, 0, 1:-1].tolist())  # 10
    edges.extend(mi[-1, -1, 1:-1].tolist())  # 11
    edges.extend(mi[0, -1, 1:-1].tolist())  # 12

    # faces - this one is extra funky
    faces = []
    faces.extend(mi[1:-1, -2:0:-1, 0].tolist())  # 1
    faces.extend(mi[1:-1, 0, 1:-1].tolist())  # 2
    faces.extend(mi[-1, 1:-1, 1:-1].tolist())  # 3
    faces.extend(mi[-2:0:-1, -1, 1:-1].tolist())  # 4
    faces.extend(mi[0, -2:0:-1, 1:-1].tolist())  # 5
    faces.extend(mi[1:-1, 1:-1, -1].tolist())  # 6

    # internal - thank you for being so friendly
    internal = mi[1:-1, 1:-1, 1:-1].tolist()

    # merge them
    to_m.extend(vertices)
    to_m.extend(edges)
    to_m.extend(faces)
    to_m.extend(internal)

    return to_m, _np.argsort(to_m).tolist()


def dof_mapping(spline):
    """
    MFEM stores vertices in a special order. This function gives dof index
    mapping between splinepy and mfem

    Parameters
    ----------
    spline: Spline

    Returns
    -------
    to_mfem: list
    from_mfem: array
    """
    if spline.para_dim == 2:
        return _mfem_dof_map_2d(spline)
    elif spline.para_dim == 3:
        return _mfem_dof_map_3d(spline)
    else:
        raise ValueError("supports para_dim of 2 and 3")


def export_cartesian(
    fname,
    spline_list,
    tolerance=None,
):
    """
    Export list of bezier splines in mfem export.

    Parameters
    ----------
    fname: string
      file name
    bezier_list: list
      list of bezier spline objects
    tolerance : float
      tolerance to collapse two neighboring points

    Returns
    -------
    None
    """
    from splinepy import Multipatch as _Multipatch

    # Check first spline
    if not isinstance(spline_list, (list, _Multipatch)):
        raise ValueError("export_cartesian expects list for export.")
    if isinstance(spline_list, list):
        spline_list = _Multipatch(splines=spline_list)

    # Set Tolerance
    if tolerance is None:
        tolerance = 1e-5

    # Set Problem dimensions
    para_dim = spline_list.para_dim
    dim = spline_list.dim
    if not ((para_dim == dim) and (dim in {2, 3})):
        raise ValueError("Only 2D2D or 3D3D splines are supported")

    # Auxiliary function to identify corner vertices of the underlying spline
    #  representation. As this function requires the numeration of its
    # underlying hypercube Element, the following part needs to be different
    # for the 2D and the 3D case
    if para_dim == 2:
        geometry_type = 3
        boundary_type = 1
        n_vertex_per_element = 4
        n_vertex_per_boundary = 2

        def _corner_vertex_ids(spline):
            cmr = spline.control_mesh_resolutions
            return _np.array(
                [
                    0,
                    cmr[0] - 1,
                    cmr[0] * cmr[1] - 1,
                    cmr[0] * (cmr[1] - 1),
                ],
                dtype="int32",
            )

        # Face enumeration is different from splinepy's
        sub_element_vertices = _np.array([[0, 1], [1, 2], [3, 2], [0, 3]])
        # splinepy's face 3 corresponds to mfem's face 0
        splinepy_face_id_2_mfem_face_id = _np.array([3, 1, 0, 2])

    elif para_dim == 3:
        geometry_type = 5
        boundary_type = 3
        n_vertex_per_element = 8
        n_vertex_per_boundary = 4

        def _corner_vertex_ids(spline):
            cmr = spline.control_mesh_resolutions
            return _np.array(
                [
                    0,
                    cmr[0] - 1,
                    cmr[0] * cmr[1] - 1,
                    cmr[0] * (cmr[1] - 1),
                    (cmr[2] - 1) * cmr[1] * cmr[0],
                    (cmr[2] - 1) * cmr[1] * cmr[0] + cmr[0] - 1,
                    (cmr[2] - 1) * cmr[1] * cmr[0] + cmr[0] * cmr[1] - 1,
                    (cmr[2] - 1) * cmr[1] * cmr[0] + cmr[0] * (cmr[1] - 1),
                ],
                dtype="int32",
            )

        sub_element_vertices = _np.array(
            [
                [0, 1, 2, 3],
                [1, 2, 6, 5],
                [3, 2, 6, 7],
                [0, 3, 7, 4],
                [0, 1, 5, 4],
                [4, 5, 6, 7],
            ]
        )
        splinepy_face_id_2_mfem_face_id = _np.array([3, 1, 4, 2, 0, 5])

    # Create a list of all corner vertices ordered by spline patch number
    corner_vertices = _np.vstack(
        [
            spline.cps[_corner_vertex_ids(spline), :]
            for spline in spline_list.patches
        ]
    )

    # Retrieve information using bezman
    connectivity = spline_list.interfaces

    (_, _, inverse_numeration, _) = _gus.utils.arr.close_rows(
        corner_vertices, tolerance, return_intersection=False
    )
    vertex_ids = inverse_numeration.reshape(-1, n_vertex_per_element)
    # Get boundaries from interfaces
    boundary_elements, boundary_faces = _np.where(connectivity < 0)
    boundary_ids = -connectivity[boundary_elements, boundary_faces]
    # Convert from splinepy enumeration to mfem enumeration
    boundaries = _np.take_along_axis(
        vertex_ids[boundary_elements, :],
        sub_element_vertices[splinepy_face_id_2_mfem_face_id[boundary_faces]],
        axis=1,
    )

    # Write all gathered information into a file
    with open(fname, "w") as f:
        # Header
        f.write(f"MFEM NURBS mesh v1.0\n\ndimension\n{dim}\n\n")

        # Elements
        n_elements = len(spline_list.patches)
        f.write(f"elements\n{n_elements}\n")
        f.write(
            "\n".join(
                f"1 {geometry_type} " + " ".join(str(id) for id in row)
                for row in vertex_ids.reshape(
                    -1, n_vertex_per_element
                ).tolist()
            )
        )

        # Boundaries
        n_boundaries = boundaries.shape[0]
        f.write(f"\n\nboundary\n{n_boundaries}\n")
        # Here currently all boundaries are set to 1
        f.write(
            "\n".join(
                f"{boundary_id} {boundary_type} "
                + " ".join(str(id) for id in row)
                for row, boundary_id in zip(
                    boundaries.reshape(-1, n_vertex_per_boundary).tolist(),
                    boundary_ids.tolist(),
                )
            )
        )

        # Write number of edges 0, for auto knot2edge
        f.write(f"\n\nedges\n{0}\n")

        # Write Number Of vertices
        f.write(f"\nvertices\n{int(_np.max(vertex_ids)+1)}\n\n")

        # Export Splines
        f.write("patches\n\n")
        for spline in spline_list.patches:
            f.write(f"knotvectors\n{para_dim}\n")
            cmr = spline.control_mesh_resolutions
            for i_para_dim in range(para_dim):
                f.write(f"{spline.degrees[i_para_dim]} ")
                f.write(f"{cmr[i_para_dim]} ")
                if "knot_vectors" not in spline.required_properties:
                    f.write("0.0 " * int(spline.degrees[i_para_dim] + 1))
                    f.write(
                        "1.0 " * int(spline.degrees[i_para_dim] + 1) + "\n"
                    )
                else:
                    f.write(
                        " ".join(str(kvi) for kvi in spline.kvs[i_para_dim])
                        + "\n"
                    )
            f.write(f"dimension\n{dim}\n")
            f.write("controlpoints_cartesian\n")
            if "weights" not in spline.required_properties:
                f.write(
                    "\n".join(
                        (" ".join(str(x_i) for x_i in row) + " 1.0")
                        for row in spline.control_points.tolist()
                    )
                    + "\n\n"
                )
            else:
                f.write(
                    "\n".join(
                        (
                            " ".join(str(x_i) for x_i in coords)
                            + " "
                            + str(weight[0])
                        )
                        for (coords, weight) in zip(
                            spline.control_points.tolist(),
                            spline.weights.tolist(),
                        )
                    )
                    + "\n\n"
                )

        f.close()


def export(fname, nurbs, precision=10):
    """Exports current nurbs in `mfem` format.

    IDs of MFEM Geometry types are:

    - Segment = 1
    - Square = 3
    - CUBS = 5

    Sections required are:

    - dimension  (space dim)
    - elements
    - boundary
    - edges
    - vertices
    - knotvectors
    - weights
    - FiniteElementSpace
    - FiniteElementCollection: <name>
    - VDim
    - Ordering (1?)

    Parameters
    -----------
    nurbs: NURBS
    fname: str

    Returns
    --------
    None
    """
    if nurbs.whatami.startswith("NURBS"):
        pass
    else:
        nurbs = nurbs.nurbs  # turns it into nurbs

    intro_sec = _form_lines(
        "MFEM NURBS mesh v1.0",
        "",
        "#",
        "# Generated with splinepy",
        "#",
        "# MFEM Geometry Types (see mesh/geom.hpp)",
        "#",
        "# SEGMENT = 1",
        "# SQUARE  = 3",
        "# CUBE    = 5",
        "#",
        "",
    )

    dimension_sec = _form_lines(
        "dimension",
        str(nurbs.para_dim),
        "",
    )

    if nurbs.para_dim == 2:
        elements_sec = _form_lines(
            "elements",
            "1",
            "1 3 0 1 2 3",
            "",
        )

        boundary_sec = _form_lines(
            "boundary",
            "4",
            "1 1 0 1",
            "2 1 2 3",
            "3 1 3 0",
            "4 1 1 2",
            "",
        )

        edges_sec = _form_lines(
            "edges",
            "4",
            "0 0 1",
            "0 3 2",
            "1 0 3",
            "1 1 2",
            "",
        )

        vertices_sec = _form_lines(
            "vertices",
            "4",
            "",
        )

        # I am not sure if mixed order is allowed, but in case not, let's
        # match orders
        max_degree = max(nurbs.degrees)
        for i, d in enumerate(nurbs.degrees):
            d_diff = int(max_degree - d)
            if d_diff > 0:
                for _ in range(d_diff):
                    nurbs.elevate_degrees(i)

        cnr = nurbs.control_mesh_resolutions

        # double-check
        if not (nurbs.degrees == nurbs.degrees[0]).all():
            raise RuntimeError(
                "Something went wrong trying to match degrees of nurbs "
                + "before export."
            )

        # This is reusable
        def kv_sec(spline):
            kvs = _form_lines(
                "knotvectors",
                str(len(spline.knot_vectors)),
            )
            kvs2 = ""
            for i in range(spline.para_dim):
                kvs2 += _form_lines(
                    str(spline.degrees[i])
                    + " "
                    + str(cnr[i])
                    + " "
                    + str(list(spline.knot_vectors[i]))[1:-1].replace(",", "")
                )

            kvs = kvs + kvs2

            kvs += "\n"  # missing empty line

            return kvs

        knotvectors_sec = kv_sec(nurbs)

    else:
        raise NotImplementedError

    # disregard inverse
    reorder_ids, _ = dof_mapping(nurbs)

    with _np.printoptions(
        formatter={"float_kind": lambda x: f"{x:.{precision}f}"}
    ):
        # weights - string operation
        weights_sec = str(nurbs.weights.flatten()[reorder_ids].tolist())
        weights_sec = weights_sec[1:-1]  # remove []
        weights_sec = weights_sec.replace("\n", "")  # remove \n
        weights_sec = weights_sec.replace(",", "")  # remove ,
        weights_sec = weights_sec.replace(" ", "\n")  # add \n
        weights_sec = "weights\n" + weights_sec  # add title
        weights_sec += "\n\n"  # empty line

        # cps
        cps_sec = str(nurbs.control_points[reorder_ids].tolist())
        cps_sec = cps_sec.replace("[", "")  # remove [
        cps_sec = cps_sec.replace(",", "")  # remove ,
        cps_sec = cps_sec.replace("\n", "")  # remove \n
        cps_sec = cps_sec.replace("]", "\n")  # replace ] with \n

    fe_space_sec = _form_lines(
        "FiniteElementSpace",
        "FiniteElementCollection: NURBS" + str(nurbs.degrees[0]),
        "VDim: " + str(nurbs.dim),
        "Ordering: 1",
        "",
    )

    with open(fname, "w") as f:
        f.write(intro_sec)
        f.write(dimension_sec)
        f.write(elements_sec)
        f.write(boundary_sec)
        f.write(edges_sec)
        f.write(vertices_sec)
        f.write(knotvectors_sec)
        f.write(weights_sec)
        f.write(fe_space_sec)
        f.write(cps_sec)



import numpy as np

import splinepy

if __name__ == "__main__":
    kv = [[0, 0, 0, 0.5, 1, 1, 1], [0, 0, 0, 1, 1, 1], [0, 0, 1, 1]]
    cp1 = np.array(
        [
            [0, 0, 0],
            [0, 1, 0],
            [1, 1.5, 0],
            [3, 1.5, 0],
            [-1, 0, 0],
            [-1, 2, 0],
            [1, 4, 0],
            [3, 4, 0],
            [-2, 0, 0],
            [-2, 2, 0],
            [1, 5, 0],
            [3, 5, 2],
        ]
    )

    cp1 = np.vstack((cp1, cp1 + [0, 0, 1]))

    # generate splines
    b = splinepy.BSpline(
        degrees=[2, 2, 1],
        knot_vectors=kv,
        control_points=cp1,
    )
    b.insert_knots(0, [0.1, 0.2, 0.3, 0.4, 0.6, 0.7, 0.8, 0.9])
    b.insert_knots(1, [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9])
    b.insert_knots(2, [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9])

    n = b.nurbs
    # edit a weight so that the values aren't the same
    n.weights[0, 0] = 0.5  # this is a tall array

    # queries
    q = [
        [0.01, 0.01, 0.01],
        # [.2, .3, .4],
        # [.3, .4, .5],
        # [.4, .5, .6],
        [0.51, 0.61, 0.71],
    ]

    print("BSpline basis functions and support ids:")
    print(b.basis_and_support(q))

    print("NURBS basis functions and support ids:")
    print(n.basis_and_support(q))

    print("These can also be stored within a matrix (scipy/numpy)")
    basis_function_matrix = splinepy.utils.data.make_matrix(
        *b.basis_and_support(q), b.control_points.shape[0], as_array=True
    )
    print(basis_function_matrix)



import numpy as _np

from splinepy.bezier import Bezier as _Bezier
from splinepy.microstructure.tiles.tile_base import TileBase as _TileBase


class Cross3DLinear(_TileBase):
    """Simple crosstile with tri-linear branches and a tri-linear
    center spline.

    .. raw:: html

        <p><a href="../_static/Cross3DLinear.html">Fullscreen</a>.</p>
        <embed type="text/html" width="100%" height="400" src="../_static/Cross3DLinear.html" />

    """  # noqa: E501

    _dim = 3
    _para_dim = 3
    _evaluation_points = _np.array(
        [
            [0.0, 0.5, 0.5],
            [1.0, 0.5, 0.5],
            [0.5, 0.0, 0.5],
            [0.5, 1.0, 0.5],
            [0.5, 0.5, 0.0],
            [0.5, 0.5, 1.0],
        ]
    )
    _n_info_per_eval_point = 1

    def _closing_tile(
        self,
        parameters=None,
        parameter_sensitivities=None,
        closure=None,
        boundary_width=0.1,
        filling_height=0.5,
        **kwargs,  # noqa ARG002
    ):
        """Create a closing tile to match with closed surface.

        Parameters
        ----------
        parameters : np.ndarray(6, 1)
          Six evaluation points with one parameter is used. This parameter
          describes the radius of the cylinder at the evaluation point.
          The parameters must be a two-dimensional np.array, where the
          value must be between 0.01 and 0.49
        parameter_sensitivities: np.ndarray(6, 1, para_dim)
          Describes the parameter sensitivities with respect to some design
          variable. In case the design variables directly apply to the
          parameter itself, they evaluate as delta_ij.
        boundary_width : float
          with of the boundary surrounding branch
        filling_height : float
          portion of the height that is filled in parametric domain
        closure : str
          parametric dimension that needs to be closed.
          Must be {"z_min", "z_max"}

        Returns
        -------
        list_of_splines : list
        derivative_list : list / None
        """
        # Check parameters
        if closure is None:
            raise ValueError("No closing direction given")

        if parameters is None:
            self._logd("Tile request is not parametrized, setting default 0.2")
            parameters = _np.array(
                _np.ones(
                    (len(self._evaluation_points), self._n_info_per_eval_point)
                )
                * 0.2
            )

        if not (_np.all(parameters > 0) and _np.all(parameters < 0.5)):
            raise ValueError("Thickness out of range (0, .5)")

        # Check if user requests derivative splines
        if parameter_sensitivities is not None:
            self.check_param_derivatives(parameter_sensitivities)
            n_derivatives = parameter_sensitivities.shape[2]
            derivatives = []
        else:
            n_derivatives = 0
            derivatives = None

        if not (0.0 < float(boundary_width) < 0.5):
            raise ValueError("Boundary Width is out of range")

        if not (0.0 < float(filling_height) < 1.0):
            raise ValueError("Filling must  be in (0,1)")

        splines = []
        for i_derivative in range(n_derivatives + 1):
            if i_derivative == 0:
                bound_width = boundary_width
                fill_height = filling_height
                inv_boundary_width = 1.0 - boundary_width
                inv_filling_height = 1.0 - filling_height
                center_width = 1.0 - (2 * boundary_width)
                r_center = center_width * 0.5
                v_zero = 0.0
                v_one_half = 0.5
                v_one = 1.0
            else:
                bound_width = 0.0
                fill_height = 0.0
                inv_boundary_width = 0.0
                inv_filling_height = 0.0
                center_width = 0.0
                r_center = center_width * 0.0
                v_zero = 0.0
                v_one_half = 0.0
                v_one = 0.0

            spline_list = []
            if closure == "z_min":
                branch_thickness = (
                    parameters[5, 0]
                    if i_derivative == 0
                    else parameter_sensitivities[5, 0, i_derivative - 1]
                )

                ctps_corner = _np.array(
                    [
                        [v_zero, v_zero, v_zero],
                        [bound_width, v_zero, v_zero],
                        [v_zero, bound_width, v_zero],
                        [bound_width, bound_width, v_zero],
                        [v_zero, v_zero, fill_height],
                        [bound_width, v_zero, fill_height],
                        [v_zero, bound_width, fill_height],
                        [bound_width, bound_width, fill_height],
                    ]
                )
                spline_list.append(
                    _Bezier(degrees=[1, 1, 1], control_points=ctps_corner)
                )

                spline_list.append(
                    _Bezier(
                        degrees=[1, 1, 1],
                        control_points=(
                            ctps_corner
                            + _np.array([v_zero, inv_boundary_width, v_zero])
                        ),
                    )
                )

                spline_list.append(
                    _Bezier(
                        degrees=[1, 1, 1],
                        control_points=(
                            ctps_corner
                            + _np.array([inv_boundary_width, v_zero, v_zero])
                        ),
                    )
                )

                spline_list.append(
                    _Bezier(
                        degrees=[1, 1, 1],
                        control_points=(
                            ctps_corner
                            + _np.array(
                                [
                                    inv_boundary_width,
                                    inv_boundary_width,
                                    v_zero,
                                ]
                            )
                        ),
                    )
                )

                center_ctps = _np.array(
                    [
                        [bound_width, bound_width, v_zero],
                        [inv_boundary_width, bound_width, v_zero],
                        [bound_width, inv_boundary_width, v_zero],
                        [inv_boundary_width, inv_boundary_width, v_zero],
                        [bound_width, bound_width, fill_height],
                        [inv_boundary_width, bound_width, fill_height],
                        [bound_width, inv_boundary_width, fill_height],
                        [inv_boundary_width, inv_boundary_width, fill_height],
                    ]
                )
                spline_list.append(
                    _Bezier(degrees=[1, 1, 1], control_points=center_ctps)
                )

                spline_list.append(
                    _Bezier(
                        degrees=[1, 1, 1],
                        control_points=(
                            _np.maximum(
                                center_ctps
                                - _np.array([center_width, v_zero, v_zero]),
                                v_zero,
                            )
                            if i_derivative == 0
                            else _np.zeros((8, 3))
                        ),
                    )
                )

                spline_list.append(
                    _Bezier(
                        degrees=[1, 1, 1],
                        control_points=(
                            _np.maximum(
                                center_ctps
                                - _np.array([v_zero, center_width, v_zero]),
                                v_zero,
                            )
                            if i_derivative == 0
                            else _np.zeros((8, 3))
                        ),
                    )
                )

                spline_list.append(
                    _Bezier(
                        degrees=[1, 1, 1],
                        control_points=(
                            _np.minimum(
                                center_ctps
                                + _np.array([center_width, v_zero, v_zero]),
                                v_one,
                            )
                            if i_derivative == 0
                            else _np.zeros((8, 3))
                        ),
                    )
                )

                spline_list.append(
                    _Bezier(
                        degrees=[1, 1, 1],
                        control_points=(
                            _np.minimum(
                                center_ctps
                                + _np.array([v_zero, center_width, v_zero]),
                                v_one,
                            )
                            if i_derivative == 0
                            else _np.zeros((8, 3))
                        ),
                    )
                )

                branch_ctps = _np.array(
                    [
                        [-r_center, -r_center, fill_height],
                        [r_center, -r_center, fill_height],
                        [-r_center, r_center, fill_height],
                        [r_center, r_center, fill_height],
                        [-branch_thickness, -branch_thickness, v_one],
                        [branch_thickness, -branch_thickness, v_one],
                        [-branch_thickness, branch_thickness, v_one],
                        [branch_thickness, branch_thickness, v_one],
                    ]
                ) + _np.array([v_one_half, v_one_half, v_zero])
                spline_list.append(
                    _Bezier(degrees=[1, 1, 1], control_points=branch_ctps)
                )
            elif closure == "z_max":
                branch_thickness = (
                    parameters[4, 0]
                    if i_derivative == 0
                    else parameter_sensitivities[4, 0, i_derivative - 1]
                )
                ctps_corner = _np.array(
                    [
                        [v_zero, v_zero, inv_filling_height],
                        [bound_width, v_zero, inv_filling_height],
                        [v_zero, bound_width, inv_filling_height],
                        [bound_width, bound_width, inv_filling_height],
                        [v_zero, v_zero, v_one],
                        [bound_width, v_zero, v_one],
                        [v_zero, bound_width, v_one],
                        [bound_width, bound_width, v_one],
                    ]
                )

                spline_list.append(
                    _Bezier(degrees=[1, 1, 1], control_points=ctps_corner)
                )

                spline_list.append(
                    _Bezier(
                        degrees=[1, 1, 1],
                        control_points=(
                            ctps_corner
                            + _np.array([v_zero, inv_boundary_width, v_zero])
                        ),
                    )
                )

                spline_list.append(
                    _Bezier(
                        degrees=[1, 1, 1],
                        control_points=(
                            ctps_corner
                            + _np.array([inv_boundary_width, v_zero, v_zero])
                        ),
                    )
                )

                spline_list.append(
                    _Bezier(
                        degrees=[1, 1, 1],
                        control_points=(
                            ctps_corner
                            + _np.array(
                                [
                                    inv_boundary_width,
                                    inv_boundary_width,
                                    v_zero,
                                ]
                            )
                        ),
                    )
                )

                center_ctps = _np.array(
                    [
                        [bound_width, bound_width, inv_filling_height],
                        [inv_boundary_width, bound_width, inv_filling_height],
                        [bound_width, inv_boundary_width, inv_filling_height],
                        [
                            inv_boundary_width,
                            inv_boundary_width,
                            inv_filling_height,
                        ],
                        [bound_width, bound_width, v_one],
                        [inv_boundary_width, bound_width, v_one],
                        [bound_width, inv_boundary_width, v_one],
                        [inv_boundary_width, inv_boundary_width, v_one],
                    ]
                )
                spline_list.append(
                    _Bezier(degrees=[1, 1, 1], control_points=center_ctps)
                )

                spline_list.append(
                    _Bezier(
                        degrees=[1, 1, 1],
                        control_points=(
                            _np.maximum(
                                center_ctps
                                - _np.array([center_width, v_zero, v_zero]),
                                v_zero,
                            )
                            if i_derivative == 0
                            else _np.zeros((8, 3))
                        ),
                    )
                )

                spline_list.append(
                    _Bezier(
                        degrees=[1, 1, 1],
                        control_points=(
                            _np.maximum(
                                center_ctps
                                - _np.array([v_zero, center_width, v_zero]),
                                v_zero,
                            )
                            if i_derivative == 0
                            else _np.zeros((8, 3))
                        ),
                    )
                )

                spline_list.append(
                    _Bezier(
                        degrees=[1, 1, 1],
                        control_points=(
                            _np.minimum(
                                center_ctps
                                + _np.array([center_width, v_zero, v_zero]),
                                v_one,
                            )
                            if i_derivative == 0
                            else _np.zeros((8, 3))
                        ),
                    )
                )

                spline_list.append(
                    _Bezier(
                        degrees=[1, 1, 1],
                        control_points=(
                            _np.minimum(
                                center_ctps
                                + _np.array([v_zero, center_width, v_zero]),
                                v_one,
                            )
                            if i_derivative == 0
                            else _np.zeros((8, 3))
                        ),
                    )
                )

                branch_ctps = _np.array(
                    [
                        [-branch_thickness, -branch_thickness, v_zero],
                        [branch_thickness, -branch_thickness, v_zero],
                        [-branch_thickness, branch_thickness, v_zero],
                        [branch_thickness, branch_thickness, v_zero],
                        [-r_center, -r_center, inv_filling_height],
                        [r_center, -r_center, inv_filling_height],
                        [-r_center, r_center, inv_filling_height],
                        [r_center, r_center, inv_filling_height],
                    ]
                ) + _np.array([v_one_half, v_one_half, v_zero])
                spline_list.append(
                    _Bezier(degrees=[1, 1, 1], control_points=branch_ctps)
                )
            else:
                raise NotImplementedError(
                    "Requested closing dimension is not supported"
                )

            if i_derivative == 0:
                splines = spline_list.copy()
            else:
                derivatives.append(spline_list)

        return (splines, derivatives)

    def create_tile(
        self,
        parameters=None,
        parameter_sensitivities=None,
        center_expansion=1.0,
        closure=None,
        **kwargs,  # noqa ARG002
    ):
        """Create a microtile based on the parameters that describe the branch
        thicknesses.

        Thickness parameters are used to describe the inner radius of the
        outward facing branches

        Parameters
        ----------
        parameters : np.array(6, 1)
          Six evaluation points with one parameter is used. This parameter
          describes the radius of the cylinder at the evaluation point.
          The parameters must be a two-dimensional np.array, where the
          value must be between 0.01 and 0.49
        parameter_sensitivities: np.ndarray(6, 1, para_dim)
          Describes the parameter sensitivities with respect to some design
          variable. In case the design variables directly apply to the
          parameter itself, they evaluate as delta_ij.
        center_expansion : float
          thickness of center is expanded by a factor
        closure : str
          parametric dimension that needs to be closed.
          Must be {"z_min", "z_max"}
        **kwargs
          Will be passed to _closing_tile if given

        Returns
        -------
        microtile_list : list(splines)
        derivative_list : list / None
        """

        if not isinstance(center_expansion, float):
            raise ValueError("Invalid Type")

        if not ((center_expansion > 0.5) and (center_expansion < 1.5)):
            raise ValueError("Center Expansion must be in (.5,1.5)")

        # Max radius, so there is no tanglement in the crosstile
        max_radius = min(0.5, (0.5 / center_expansion))

        # set to default if nothing is given
        if parameters is None:
            self._logd("Setting branch thickness to default 0.2")
            parameters = (
                _np.ones(
                    (
                        self._evaluation_points.shape[0],
                        self._n_info_per_eval_point,
                    )
                )
                * 0.2
            )

        # Check for type and consistency
        self.check_params(parameters)
        if _np.any(parameters <= 0) or _np.any(parameters > max_radius):
            raise ValueError(
                f"Radii must be in (0,{max_radius}) for "
                f"center_expansion {center_expansion}"
            )

        if parameter_sensitivities is not None:
            self.check_param_derivatives(parameter_sensitivities)
            n_derivatives = parameter_sensitivities.shape[2]
            derivatives = []
        else:
            n_derivatives = 0
            derivatives = None

        if closure is not None:
            return self._closing_tile(
                parameters=parameters,
                parameter_sensitivities=parameter_sensitivities,
                closure=closure,
                **kwargs,
            )

        splines = []
        for i_derivative in range(n_derivatives + 1):
            # Constant auxiliary values
            if i_derivative == 0:
                [x_min_r, x_max_r, y_min_r, y_max_r, z_min_r, z_max_r] = (
                    parameters[:, 0]
                )
                v_one_half = 0.5
                center_r = center_expansion * _np.mean(parameters[:, 0])
            else:

                [x_min_r, x_max_r, y_min_r, y_max_r, z_min_r, z_max_r] = (
                    parameter_sensitivities[:, :, i_derivative - 1].flatten()
                )
                v_one_half = 0.0
                center_r = center_expansion * _np.mean(
                    parameter_sensitivities[:, :, i_derivative - 1].flatten()
                )

            spline_list, center = [], [v_one_half] * 3

            # Create the center-tile
            center_points = _np.array(
                [
                    [-center_r, -center_r, -center_r],
                    [center_r, -center_r, -center_r],
                    [-center_r, center_r, -center_r],
                    [center_r, center_r, -center_r],
                    [-center_r, -center_r, center_r],
                    [center_r, -center_r, center_r],
                    [-center_r, center_r, center_r],
                    [center_r, center_r, center_r],
                ]
            )
            spline_list.append(
                _Bezier(
                    degrees=[1, 1, 1], control_points=center_points + center
                )
            )

            # X-Axis branches
            # X-Min-Branch
            x_min_ctps = _np.array(
                [
                    [-v_one_half, -x_min_r, -x_min_r],
                    center_points[0, :],
                    [-v_one_half, x_min_r, -x_min_r],
                    center_points[2, :],
                    [-v_one_half, -x_min_r, x_min_r],
                    center_points[4, :],
                    [-v_one_half, x_min_r, x_min_r],
                    center_points[6, :],
                ]
            )
            spline_list.append(
                _Bezier(degrees=[1, 1, 1], control_points=x_min_ctps + center)
            )

            # X-Max-Branch
            x_max_ctps = _np.array(
                [
                    center_points[1, :],
                    [v_one_half, -x_max_r, -x_max_r],
                    center_points[3, :],
                    [v_one_half, x_max_r, -x_max_r],
                    center_points[5, :],
                    [v_one_half, -x_max_r, x_max_r],
                    center_points[7, :],
                    [v_one_half, x_max_r, x_max_r],
                ]
            )
            spline_list.append(
                _Bezier(degrees=[1, 1, 1], control_points=x_max_ctps + center)
            )

            # Y-Axis branches
            # Y-Min-Branch
            y_min_ctps = _np.array(
                [
                    [-y_min_r, -v_one_half, -y_min_r],
                    [y_min_r, -v_one_half, -y_min_r],
                    center_points[0, :],
                    center_points[1, :],
                    [-y_min_r, -v_one_half, y_min_r],
                    [y_min_r, -v_one_half, y_min_r],
                    center_points[4, :],
                    center_points[5, :],
                ]
            )
            spline_list.append(
                _Bezier(degrees=[1, 1, 1], control_points=y_min_ctps + center)
            )

            # Y-Max-Branch
            y_max_ctps = _np.array(
                [
                    center_points[2, :],
                    center_points[3, :],
                    [-y_max_r, v_one_half, -y_max_r],
                    [y_max_r, v_one_half, -y_max_r],
                    center_points[6, :],
                    center_points[7, :],
                    [-y_max_r, v_one_half, y_max_r],
                    [y_max_r, v_one_half, y_max_r],
                ]
            )
            spline_list.append(
                _Bezier(degrees=[1, 1, 1], control_points=y_max_ctps + center)
            )

            # Z-Axis branches
            # Z-Min-Branch
            z_min_ctps = _np.array(
                [
                    [-z_min_r, -z_min_r, -v_one_half],
                    [z_min_r, -z_min_r, -v_one_half],
                    [-z_min_r, z_min_r, -v_one_half],
                    [z_min_r, z_min_r, -v_one_half],
                    center_points[0, :],
                    center_points[1, :],
                    center_points[2, :],
                    center_points[3, :],
                ]
            )
            spline_list.append(
                _Bezier(degrees=[1, 1, 1], control_points=z_min_ctps + center)
            )

            # Z-Max-Branch
            z_max_ctps = _np.array(
                [
                    center_points[4, :],
                    center_points[5, :],
                    center_points[6, :],
                    center_points[7, :],
                    [-z_max_r, -z_max_r, v_one_half],
                    [z_max_r, -z_max_r, v_one_half],
                    [-z_max_r, z_max_r, v_one_half],
                    [z_max_r, z_max_r, v_one_half],
                ]
            )
            spline_list.append(
                _Bezier(degrees=[1, 1, 1], control_points=z_max_ctps + center)
            )

            if i_derivative == 0:
                splines = spline_list.copy()
            else:
                derivatives.append(spline_list)

        return (splines, derivatives)



from functools import wraps as _wraps

import numpy as _np
from gustaf.utils import arr as _arr

from splinepy import settings as _settings
from splinepy.utils import log as _log
from splinepy.utils.data import make_matrix as _make_matrix


def embedded(spline, new_dimension):
    """Embeds Spline in given dimension.
    If dim > current_dim, it will add zeros columns to the control points.
    If dim < current_dim, it takes first n-columns of the control points.
    If dim == current_dim, it returns a copy.

    Parameters
    ----------
    spline: Spline
      (`self`-argument if called via extract member of a spline)
    new_dimension : integer
      new physical dimension into which the spline control points will be set

    Returns
    -------
    embedded_spline : Spline
    """
    from splinepy.spline import Spline as _Spline

    # Check input type
    if not isinstance(spline, _Spline):
        raise NotImplementedError("Embedded only works for splines")

    if new_dimension < 1:
        raise ValueError("Cannot reduce dimension below 1")

    if spline.dim > new_dimension:
        spline_dict = spline.todict()
        spline_dict["control_points"] = spline_dict["control_points"][
            :, :new_dimension
        ]
        return type(spline)(**spline_dict)
    elif spline.dim < new_dimension:
        spline_dict = spline.todict()
        spline_dict["control_points"] = _np.hstack(
            (
                spline_dict["control_points"][:, :new_dimension],
                _np.zeros((spline.cps.shape[0], new_dimension - spline.dim)),
            )
        )
        return type(spline)(**spline_dict)
    else:
        # use todict() to avoid unnecessary data copies from `copy()`
        return type(spline)(**spline.todict())


def extruded(spline, extrusion_vector=None):
    """Extrudes Splines.

    Linear Extrusion of a given spline along an extrusion vector. This will
    extend the parametric dimensionality by one. If the dimension of the
    extrusion vector is higher than the spline's dimension, the spline will be
    extended into the new dimension.

    Parameters
    ----------
    spline: Spline
      (`self`-argument if called via extract member of a spline)
    extrusion_vector: np.ndarray

    Returns
    -------
    extruded_spline : Spline
    """
    from splinepy.spline import Spline as _Spline

    # Check input type
    if not isinstance(spline, _Spline):
        raise NotImplementedError("Extrude only works for splines")

    # Check extrusion_vector
    if extrusion_vector is not None:
        # make flat extrusion_vector
        extrusion_vector = _np.asarray(extrusion_vector).ravel()
    else:
        raise ValueError("No extrusion extrusion_vector given")

    # Check extrusion_vector dimension
    # formulate correct cps
    if spline.dim == extrusion_vector.shape[0]:
        cps = spline.control_points
    elif spline.dim < extrusion_vector.shape[0]:
        expansion_dimension = extrusion_vector.shape[0] - spline.dim
        # one smaller dim is allowed
        # warn that we assume new dim is all zero
        _log.debug(
            f"Given extrusion vector is {expansion_dimension} dimension "
            "bigger than spline's dim. Assuming 0.0 entries for "
            "new dimension.",
        )
        cps = _np.hstack(
            (
                spline.control_points,
                _np.zeros((len(spline.control_points), expansion_dimension)),
            )
        )
    else:
        raise ValueError(
            "Dimension Mismatch between extrusion extrusion vector "
            "and spline."
        )

    # Start Extrusion
    spline_dict = {}

    spline_dict["degrees"] = _np.concatenate((spline.degrees, [1]))
    spline_dict["control_points"] = _np.vstack((cps, cps + extrusion_vector))
    if spline.has_knot_vectors:
        spline_dict["knot_vectors"] = spline.knot_vectors + [[0, 0, 1, 1]]
    if spline.is_rational:
        spline_dict["weights"] = _np.concatenate(
            (spline.weights, spline.weights)
        )

    return type(spline)(**spline_dict)


def revolved(
    spline, axis=None, center=None, angle=None, n_knot_spans=None, degree=True
):
    """Revolve spline around an axis and extend its parametric dimension.

    Parameters
    ----------
    spline: Spline
      (`self`-argument if called via extract member of a spline)
    axis : np.ndarray
      Axis of revolution
    center : np.ndarray
      Center of revolution
    angle : float
      angle of the revolution.
    n_knot_spans : int
      number of non-zero knot-elements for result-spline (if applicable)
    degree : bool
      use degrees instead of radiant

    Returns
    -------
    spline : Spline
    """
    from splinepy.spline import Spline as _Spline

    # Check input type
    if not isinstance(spline, _Spline):
        raise NotImplementedError("Revolutions only works for splines")

    # Check axis
    if axis is not None:
        # Transform into numpy array
        axis = _np.asarray(axis).ravel()
        # Check Axis dimension
        if spline.control_points.shape[1] > axis.shape[0]:
            raise ValueError(
                "Dimension Mismatch between extrusion axis and spline."
            )
        elif spline.control_points.shape[1] < axis.shape[0]:
            _log.debug(
                "Control Point dimension is smaller than axis dimension,"
                " filling with zeros"
            )
            expansion_dimension = axis.shape[0] - spline.dim
            cps = _np.hstack(
                (
                    spline.control_points,
                    _np.zeros(
                        (len(spline.control_points), expansion_dimension)
                    ),
                )
            )
        else:
            cps = _np.copy(spline.control_points)

        # Make sure axis is normalized

        axis_norm = _np.linalg.norm(axis)
        if not _np.isclose(axis_norm, 0, atol=_settings.TOLERANCE):
            axis = axis / axis_norm
        else:
            raise ValueError("Axis-norm is too close to zero.")
    else:
        cps = _np.copy(spline.control_points)
        if spline.control_points.shape[1] == 3:
            raise ValueError("No rotation axis given")

    # Set Problem dimension
    problem_dimension = cps.shape[1]

    # Make sure axis is ignored for 2D
    if problem_dimension == 2:
        axis = None

    # Update angle
    if angle is None:
        spline._logd("No angle given for the revolution. Using 360 degrees.")
        angle = 360 if degree else _np.pi * 2

    if degree:
        angle = _np.radians(angle)

    # Init center
    if center is not None:
        center = _np.asarray(center).ravel()
        # Check Axis dimension
        if not (problem_dimension == center.shape[0]):
            raise ValueError(
                "Dimension Mismatch between axis and center of rotation."
            )
        cps -= center

    # The parametric dimension is independent of the revolution but the
    # rotation-matrix is only implemented for 2D and 3D problems
    if cps.shape[1] not in {2, 3}:
        raise NotImplementedError(
            "Sorry," "revolutions only implemented for 2D and 3D splines"
        )

    # Angle must be (0, pi) non including
    # Rotation is always performed in half steps
    PI = _np.pi
    minimum_n_knot_spans = int(
        _np.ceil((abs(angle) + abs(_settings.TOLERANCE)) / PI)
    )
    if n_knot_spans is None or (n_knot_spans < minimum_n_knot_spans):
        n_knot_spans = minimum_n_knot_spans

    if "Bezier" in spline.name and n_knot_spans > 1:
        raise ValueError(
            "Revolutions are only supported for angles up to 180 "
            "degrees for Bezier type splines as they consist of only "
            "one knot span"
        )

    # Determine auxiliary values
    rot_a = angle / (2 * n_knot_spans)
    half_counter_angle = PI / 2 - rot_a
    weight = _np.sin(half_counter_angle)
    factor = 1 / weight

    # Determine rotation matrix
    rotation_matrix = _arr.rotation_matrix_around_axis(
        axis=axis, rotation=rot_a, degree=False
    ).T

    # Start Extrusion
    spline_dict = {}

    spline_dict["degrees"] = _np.concatenate((spline.degrees, [2]))

    spline_dict["control_points"] = cps
    end_points = cps
    for _i_segment in range(n_knot_spans):
        # Rotate around axis
        mid_points = _np.matmul(end_points, rotation_matrix)
        end_points = _np.matmul(mid_points, rotation_matrix)
        # Move away from axis using dot-product tricks
        if problem_dimension == 3:
            mp_scale = axis * _np.dot(mid_points, axis).reshape(-1, 1)
            mid_points = (mid_points - mp_scale) * factor + mp_scale
        else:
            mid_points *= factor
        spline_dict["control_points"] = _np.concatenate(
            (spline_dict["control_points"], mid_points, end_points)
        )

    if spline.has_knot_vectors:
        kv = [0, 0, 0]
        [kv.extend([i + 1, i + 1]) for i in range(n_knot_spans - 1)]
        spline_dict["knot_vectors"] = spline.knot_vectors + [
            kv + [n_knot_spans] * 3
        ]
    if spline.is_rational:
        mid_weights = spline.weights * weight
        spline_dict["weights"] = spline.weights
        for _i_segment in range(n_knot_spans):
            spline_dict["weights"] = _np.concatenate(
                (spline_dict["weights"], mid_weights, spline.weights)
            )
    else:
        _log.debug(
            "True revolutions are only possible for rational spline types.",
            "Creating Approximation.",
        )

    if center is not None:
        spline_dict["control_points"] += center

    return type(spline)(**spline_dict)


def from_bounds(parametric_bounds, physical_bounds):
    """Creates a minimal spline with given parametric bounds, physical bounds.
    Physical bounds can have less or equal number of
    dimension as parametric bounds. (Greater is not supported)

    Parameters
    -----------
    parametric_bounds: (2, n) array-like
    physical_bounds: (2, n) array-like

    Returns
    --------
    spline: BSpline
    """
    physical_bounds = _np.asanyarray(physical_bounds).reshape(2, -1)
    parametric_bounds = _np.asanyarray(parametric_bounds).reshape(2, -1)

    # get correctly sized bez box
    phys_size = physical_bounds[1] - physical_bounds[0]

    # minimal bspline
    bspline_box = box(*phys_size).bspline  # kvs are in [0, 1]
    bspline_box.cps += physical_bounds[0]  # apply offset

    # update parametric bounds
    new_kvs = []
    para_size = parametric_bounds[1] - parametric_bounds[0]
    for i, kv in enumerate(bspline_box.kvs):
        # apply scale and offset
        new_kv = (kv * para_size[i]) + parametric_bounds[0][i]
        new_kvs.append(new_kv)

    # update at once
    bspline_box.kvs = new_kvs

    return bspline_box


def determinant_spline(spline):
    """
    Creates a spline representing the projection of the jacobian determinant

    dmin > 0 --> spline is not tangled
    dmin <= 0 --> spline is COULD BE tangled

    Only works if (parameter dimension == physical dimension).
    A definitive statement about the entanglement can only be made with
    non-rational splines or rational splines with equal weights!
    Otherwise, the resulting spline, and therefore the entanglement check
    is just an approximation.

    references:
    Mantzaflaris, A., Jüttler, B., Khoromskij, B. N., & Langer, U. (2017)
    Limkilde, A., Evgrafov, A., Gravesen, J., & Mantzaflaris, A. (2021)

    Parameters
    ----------
    spline: Spline

    Returns
    -------
    determinant_projection: BSpline
      Spline which represents the jacobian determinant of the
      given spline object.
    """
    from splinepy.utils.data import has_scipy as _has_scipy

    # choose solver (numpy/scipy)
    if _has_scipy:
        from scipy.sparse.linalg import spsolve as _spsolve

        solving_function = _spsolve

    else:
        solving_function = _np.linalg.solve

    # checks if dimensions are equal.
    if spline.dim != spline.para_dim:
        raise ValueError(
            "Dimension of parameter space and physical space"
            "are not the same ({spline.whatami})!"
        )

    # checks if spline is rational and if weights differ
    # otherwise NURBS can be treated like regular BSpline

    if spline.is_rational and not _np.allclose(
        spline.weights, spline.weights.ravel()[0]
    ):
        _log.warning(
            f"Spline is rational ({spline.whatami}) with unequal weights!"
            "Result is only an approximation."
        )

    # calculate degrees of det spline
    degrees_determinant_spline = spline.degrees * spline.dim - 1
    # calculate necessary additional knot multiplicity due to degree elevation
    multiplicity_increase = degrees_determinant_spline - spline.degrees

    # knot_vectors
    if spline.has_knot_vectors:
        # initialize lists for knot vectors
        knot_vectors_determinant_spline = []

        for u_kv, kn_m, m in zip(
            spline.unique_knots,
            spline.knot_multiplicities,
            multiplicity_increase,
        ):
            # increase knot multiplicities:
            #   @ each inner knot -> mult_inc + 1
            #   @ begin & end -> mult_inc
            # since continuity at inner knots further decreases by 1
            kn_m[1:-1] += 1
            temp_knot_vector = _np.repeat(u_kv, kn_m + m)
            knot_vectors_determinant_spline.append(temp_knot_vector)

        # number of cpts
        n_control_points = _np.prod(
            [
                len(kvs_ds) - d_ds - 1
                for kvs_ds, d_ds in zip(
                    knot_vectors_determinant_spline, degrees_determinant_spline
                )
            ]
        )
        determinant_projection = _settings.NAME_TO_TYPE["BSpline"](
            degrees=degrees_determinant_spline,
            knot_vectors=knot_vectors_determinant_spline,
            control_points=_np.empty((n_control_points, 1)),
        )

    else:
        # if Bezier Spline
        n_control_points = _np.prod(degrees_determinant_spline + 1)
        determinant_projection = _settings.NAME_TO_TYPE["Bezier"](
            degrees=degrees_determinant_spline,
            control_points=_np.empty((n_control_points, 1)),
        )

    # get det(J) of spline at greville pts
    # to calculate cpts of determinant Spline
    sample_queries = determinant_projection.greville_abscissae(
        duplicate_tolerance=_settings.TOLERANCE
    )
    jacobian_determinants = _np.linalg.det(spline.jacobian(sample_queries))
    coefficient_matrix = _make_matrix(
        *determinant_projection.basis_and_support(sample_queries),
        n_control_points,
    )
    determinant_projection.control_points[:, 0] = solving_function(
        coefficient_matrix, jacobian_determinants
    )

    return determinant_projection


def parametric_view(spline, axes=True, conform=False):
    """Create parametric view of given spline. Previously called
    `naive_spline()`. Degrees are always 1 and knot multiplicity is not
    preserved. Returns BSpline, as BSpline and NURBS should look the same as
    parametric view.
    Will take shallow copy of underlying data of spline_data and show_options
    from original spline.
    However, if conforming basis is desired, set confrom=True.

    Parameters
    -----------
    spline: BSpline or NURBS
    axes: bool
      If True, will configure axes settings, it is supported.
    conform: bool
      Default is False


    Returns
    --------
    para_spline: BSpline
    """
    p_bounds = spline.parametric_bounds
    para_spline = from_bounds(
        parametric_bounds=p_bounds, physical_bounds=p_bounds
    )

    # process to create conforming para_view splines
    if conform:
        # conforming spline - rational?
        if spline.is_rational:
            para_spline = para_spline.nurbs

        # conforming spline - bezier?
        if not spline.has_knot_vectors:
            dict_spline = para_spline.todict()
            dict_spline.pop("knot_vectors")
            para_spline = type(spline)(**dict_spline)

        # conforming degrees
        d_diff = spline.degrees - para_spline.degrees
        if any(d_diff):
            # turn difference in degrees into degree query
            d_query = []
            for i, dq in enumerate(d_diff):
                d_query += [i] * dq
            para_spline.elevate_degrees(d_query)

        # process knots to insert
        if spline.has_knot_vectors:
            for i, (kv, d) in enumerate(
                zip(spline.knot_vectors, spline.degrees)
            ):
                n_repeating = int(d + 1)
                query = kv[n_repeating:-n_repeating]
                if len(query) > 0:
                    para_spline.insert_knots(i, query)

        # conforming weights
        if spline.is_rational:
            para_spline.ws[:] = spline.ws

    elif spline.has_knot_vectors:
        # in non-conform case, we can just add missing unique knots
        for i, uk in enumerate(spline.unique_knots):
            query = uk[1:-1]
            if len(query) > 0:
                para_spline.insert_knots(i, query)

    # take shallow copy
    para_spline.spline_data._saved = spline.spline_data._saved.copy()
    para_spline.show_options._options = spline.show_options._options.copy()

    if axes and "axes" in spline.show_options.valid_keys():
        # configure axes
        bs = p_bounds
        bs_diff_001 = (bs[1] - bs[0]) * 0.001
        lower_b = bs[0] - bs_diff_001
        upper_b = bs[1] + bs_diff_001
        axes_config = {
            "xtitle": "u",
            "xrange": [lower_b[0], upper_b[0]],
            "tip_size": 0,
            "xminor_ticks": 3,
            "xygrid": False,
        }
        if spline.para_dim == 2:
            axes_config.update(
                ytitle="v",
                yrange=[lower_b[1], upper_b[1]],
                yminor_ticks=3,
                yzgrid=False,
            )

        if spline.para_dim == 3:
            axes_config.update(
                ztitle="w",
                zrange=[lower_b[2], upper_b[2]],
                zminor_ticks=3,
                zxgrid=False,
            )

        para_spline.show_options["axes"] = axes_config
        # it is a view, so cps won't be realistic
        para_spline.show_options["control_points"] = False
        para_spline.show_options["lighting"] = "off"

    return para_spline


def line(points):
    """Create a spline with the provided points as control points.

    Parameters
    ----------
    points: (n, d) numpy.ndarray
      npoints x ndims array of control points

    Returns
    -------
    line: BSpline
      Spline degree [1].
    """
    # lines have degree 1
    degree = 1

    cps = _np.array(points)
    nknots = cps.shape[0] + degree + 1

    knots = _np.concatenate(
        (
            _np.full(degree, 0.0),
            _np.linspace(0.0, 1.0, nknots - 2 * degree),
            _np.full(degree, 1.0),
        )
    )

    spline = _settings.NAME_TO_TYPE["BSpline"](
        control_points=cps, knot_vectors=[knots], degrees=[degree]
    )

    return spline


def arc(
    radius=1.0,
    angle=90.0,
    n_knot_spans=-1,
    start_angle=0.0,
    degree=True,
):
    """Creates a 1-D arc as Rational Bezier or NURBS with given radius and
    angle. The arc lies in the x-y plane and rotates around the z-axis.

    Parameters
    ----------
    radius : float, optional
      radius of the arc, defaults to 1
    angle : float, optional
      angle of the section of the arc, defaults to 90 degrees
    n_knot_spans : int
      Number of knot spans, by default minimum number for angle is used.
    start_angle : float, optional
      starting point of the angle, by default 0.
    degree: bool, optional
      degrees for angle used, by default True

    Returns
    -------
    arc: NURBS or RationalBezier
    """
    # Define point spline of degree 0 at starting point of the arc
    if degree:
        start_angle = _np.radians(start_angle)
        angle = _np.radians(angle)
    start_point = [
        radius * _np.cos(start_angle),
        radius * _np.sin(start_angle),
    ]
    point_spline = _settings.NAME_TO_TYPE["RationalBezier"](
        degrees=[0], control_points=[start_point], weights=[1.0]
    )
    # Bezier splines only support angles lower than 180 degrees
    if abs(angle) >= _np.pi or n_knot_spans > 1:
        point_spline = point_spline.nurbs

    # Revolve - set degree to False, since all the angles are converted to rad
    arc_attrib = point_spline.create.revolved(
        angle=angle, n_knot_spans=n_knot_spans, degree=False
    ).todict()
    # Remove the first parametric dimensions, which is only a point and
    # only used for the revolution
    arc_attrib["degrees"] = list(arc_attrib["degrees"])[1:]
    if point_spline.has_knot_vectors:
        arc_attrib["knot_vectors"] = list(arc_attrib["knot_vectors"])[1:]

    return type(point_spline)(**arc_attrib)


def circle(radius=1.0, n_knot_spans=3):
    """Circle (parametric dim = 1) with radius r in the x-y plane around the
    origin. The spline has an open knot vector and degree 2.

    Parameters
    ----------
    radius : float, optional
      radius, defaults to one
    n_knots_spans : int, optional
      number of knot spans, defaults to 3

    Returns
    -------
    circle: NURBS
    """
    return arc(radius=radius, angle=360, n_knot_spans=n_knot_spans)


def box(*lengths):
    """N-D box (hyper rectangle).

    Parameters
    ----------
    *lengths: list(float)

    Returns
    -------
    nd_box: Bezier
    """
    # may dim check here?
    # starting point
    nd_box = _settings.NAME_TO_TYPE["Bezier"](
        degrees=[1], control_points=[[0], [lengths[0]]]
    )
    # use extrude
    for i, len_ in enumerate(lengths[1:]):
        nd_box = nd_box.create.extruded([0] * int(i + 1) + [len_])

    return nd_box


def plate(radius=1.0):
    """Creates a biquadratic 2-D spline in the shape of a plate with given
    radius.

    Parameters
    ----------
    radius : float, optional
      Radius of the plate, defaults to one

    Returns
    -------
    plate: RationalBezier
    """
    degrees = [2, 2]
    control_points = (
        _np.array(
            [
                [-0.5, -0.5],
                [0.0, -1.0],
                [0.5, -0.5],
                [-1.0, 0.0],
                [0.0, 0.0],
                [1.0, 0.0],
                [-0.5, 0.5],
                [0.0, 1.0],
                [0.5, 0.5],
            ]
        )
        * radius
    )
    weights = _np.tile([1.0, 1 / _np.sqrt(2)], 5)[:-1]

    return _settings.NAME_TO_TYPE["RationalBezier"](
        degrees=degrees, control_points=control_points, weights=weights
    )


def disk(
    outer_radius,
    inner_radius=None,
    angle=360.0,
    n_knot_spans=4,
    degree=True,
):
    """Surface spline describing a potentially hollow disk with quadratic
    degree along curved dimension and linear along thickness. The angle
    describes the returned part of the disk.

    Parameters
    ----------
    outer_radius : float
      Outer radius of the disk
    inner_radius : float, optional
      Inner radius of the disk, in case of hollow disk, by default 0.
    angle : float, optional
      Rotational angle, by default 360. describing a complete revolution
    n_knot_spans : int, optional
      Number of knot spans, by default 4

    Returns
    -------
    disk: NURBS
      Surface NURBS of degrees (1,2)
    """
    if inner_radius is None:
        inner_radius = 0.0

    cps = _np.array([[inner_radius, 0.0], [outer_radius, 0.0]])
    weights = _np.ones([cps.shape[0]])
    knots = _np.repeat([0.0, 1.0], 2)

    return _settings.NAME_TO_TYPE["NURBS"](
        control_points=cps, knot_vectors=[knots], degrees=[1], weights=weights
    ).create.revolved(angle=angle, n_knot_spans=n_knot_spans, degree=degree)


def torus(
    torus_radius,
    section_outer_radius,
    section_inner_radius=None,
    torus_angle=None,
    section_angle=None,
    section_n_knot_spans=4,
    torus_n_knot_spans=4,
    degree=True,
):
    """Creates a volumetric NURBS spline describing a torus revolved around the
    x-axis. Possible cross-sections are plate, disk (yielding a tube) and
    section of a disk.

    Parameters
    ----------
    torus_radius : float
      Radius of the torus
    section_outer_radius : float
      Radius of the section of the torus
    section_inner_radius : float, optional
      Inner radius in case of hollow torus, by default 0.
    torus_angle : float, optional
      Rotational angle of the torus, by default None, giving a complete
      revolution
    section_angle : float, optional
      Rotational angle, by default None, yielding a complete revolution
    section_n_knot_spans : float, optional
      Number of knot spans along the cross-section, by default 4
    torus_n_knot_spans : float, optional
      Number of knot spans along the torus, by default 4

    Returns
    -------
    torus: NURBS
      Volumetric spline in the shape of a torus with degrees (1,2,2)
    """

    if torus_angle is None:
        torus_angle = 2 * _np.pi
        degree = False

    if section_angle is None:
        section_angle = 2 * _np.pi
        section_angle_flag = False
        degree = False
    else:
        section_angle_flag = True

    if section_inner_radius is None:
        section_inner_radius = 0
        section_inner_radius_flag = False
    else:
        section_inner_radius_flag = True

    # Create the cross-section
    if not section_angle_flag and not section_inner_radius_flag:
        cross_section = plate(section_outer_radius)
        # For more than 180 degree only NURBS can be used
        if abs(torus_angle) >= _np.pi:
            cross_section = cross_section.nurbs
    else:
        cross_section = disk(
            outer_radius=section_outer_radius,
            inner_radius=section_inner_radius,
            n_knot_spans=section_n_knot_spans,
            angle=section_angle,
            degree=degree,
        )

    # Create a surface spline representing a disk and move it from the origin
    cross_section.control_points[:, 1] += torus_radius

    return cross_section.create.revolved(
        axis=[1.0, 0, 0],
        center=_np.zeros(3),
        angle=torus_angle,
        n_knot_spans=torus_n_knot_spans,
        degree=degree,
    )


def sphere(
    outer_radius,
    inner_radius=None,
    angle=360.0,
    n_knot_spans=-1,
    degree=True,
):
    """Creates a volumetric spline describing a sphere with radius R.

    Parameters
    ----------
    outer_radius : float
      Outer radius of the sphere
    inner_radius : float, optional
      Inner radius of the potentially hollow sphere.
    angle : float
      Rotational angle around x-axis, by default each 360
      (describing a complete revolution)
    n_knot_spans : int
      Number of knot spans

    Returns
    -------
    sphere: NURBS
      Volumetric NURBS with degrees (1,2,2)
    """
    if inner_radius is not None:
        inner_radius = float(inner_radius)

    sphere = disk(
        outer_radius, inner_radius, angle=180, degree=True
    ).nurbs.create.revolved(
        axis=[1, 0, 0],
        angle=angle,
        n_knot_spans=n_knot_spans,
        degree=degree,
    )

    return sphere


def surface_circle(outer_radius):
    """
    Create a circle consisting of a single surface patch

    Parameters
    ----------
    outer_rarius : float
      radius

    Returns
    -------
    patch : spline
      Circle spline
    """
    aux_0_w = 2**-0.5
    aux_0 = outer_radius * aux_0_w

    return _settings.NAME_TO_TYPE["RationalBezier"](
        degrees=[2, 2],
        control_points=[
            [-aux_0, -aux_0],
            [0, -2 * aux_0],
            [aux_0, -aux_0],
            [-2 * aux_0, 0],
            [0, 0],
            [2 * aux_0, 0],
            [-aux_0, aux_0],
            [0, 2 * aux_0],
            [aux_0, aux_0],
        ],
        weights=[1, aux_0_w, 1, aux_0_w, 1, aux_0_w, 1, aux_0_w, 1],
    )


def cone(
    outer_radius,
    height,
    inner_radius=None,
    volumetric=True,
    angle=360.0,
    degree=True,
):
    """Creates a cone with circular base.

    Parameters
    ----------
    radius : float
      Radius of the base
    height : float
      Height of the cone
    volumetric : bool, optional
      Parameter whether surface or volume spline, by default True
    angle : float
      Rotation angle in degrees, only used for solid model

    Returns
    -------
    cone: NURBS
      Volumetric or surface NURBS describing a cone
    """

    if volumetric:
        ground = disk(
            outer_radius, inner_radius=inner_radius, angle=angle, degree=degree
        )
    else:
        ground = circle(outer_radius)

    # Extrude in z
    cone = ground.create.extruded([0, 0, height])
    # Move all upper control points to one
    cone.control_points[_np.isclose(cone.control_points[:, -1], height)] = [
        0,
        0,
        height,
    ]

    return cone


def pyramid(width, length, height):
    """Creates a volumetric spline in the shape of a pyramid with linear degree
    in every direction.

    Parameters
    ----------
    width : float
      Dimension of base in x-axis
    length : float
      Dimension of base in y-axis
    height : float
      Height in z-direction

    Returns
    -------
    pyramid: BSpline
      Volumetric linear spline in the shape of a pyramid
    """

    # Create box
    p = box(width, length, height)

    # Collapse all upper points on one control point
    p.control_points[4:, :] = [width / 2, length / 2, height]

    return p


class Creator:
    """Helper class to build new splines from existing geometries.

    Examples
    ---------
    >>> spline_faces = my_spline.create.extrude(vector=[3, 1, 3])
    """

    __slots__ = ("_helpee",)

    def __init__(self, spl):
        self._helpee = spl

    @_wraps(extruded)
    def extruded(self, *args, **kwargs):
        return extruded(self._helpee, *args, **kwargs)

    @_wraps(embedded)
    def embedded(self, *args, **kwargs):
        return embedded(self._helpee, *args, **kwargs)

    @_wraps(revolved)
    def revolved(self, *args, **kwargs):
        return revolved(self._helpee, *args, **kwargs)

    @_wraps(parametric_view)
    def parametric_view(self, *args, **kwargs):
        return parametric_view(self._helpee, *args, **kwargs)

    @_wraps(determinant_spline)
    def determinant_spline(self, *args, **kwargs):
        return determinant_spline(self._helpee, *args, **kwargs)



import numpy as _np

from splinepy.bezier import Bezier as _Bezier
from splinepy.microstructure.tiles.tile_base import TileBase as _TileBase


class DoubleLattice(_TileBase):
    """
    Lattice base cell, consisting of a rectangle with two diagonals in the
    center.

    .. raw:: html

        <p><a href="../_static/DoubleLattice.html">Fullscreen</a>.</p>
        <embed type="text/html" width="100%" height="400" src="../_static/DoubleLattice.html" />

    """  # noqa: E501

    _dim = 2
    _para_dim = 2
    _evaluation_points = _np.array([[0.5, 0.5]])
    _n_info_per_eval_point = 2

    def create_tile(
        self,
        parameters=None,
        parameter_sensitivities=None,
        contact_length=0.5,
        **kwargs,  # noqa ARG002
    ):
        """Create a microtile based on the parameters that describe the branch
        thicknesses.

        Thickness parameters are used to describe the inner radius of the
        outward facing branches

        Parameters
        ----------
        parameters : np.array
          first entry defines the thickness of the vertical and horizontal
          branches
          second entry defines the thickness of the diagonal branches
        parameter_sensitivities: np.ndarray
          correlates with thickness of branches and entouring wall
        contact_length : double
          required for conformity between tiles, sets the length of the center
          block on the tiles boundary

        Returns
        -------
        microtile_list : list(splines)
        """
        if not isinstance(contact_length, float):
            raise ValueError("Invalid Type")
        if not ((contact_length > 0.0) and (contact_length < 1.0)):
            raise ValueError("Contact length must be in (0.,1.)")

        # set to default if nothing is given
        if parameters is None:
            self._logd("Tile request is not parametrized, setting default 0.2")
            parameters = _np.ones((1, 2)) * 0.1
        if not (
            _np.all(parameters > 0)
            and _np.all(parameters < 0.5 / (1 + _np.sqrt(2)))
        ):
            raise ValueError(
                "Parameters must be between 0.01 and 0.5/(1+sqrt(2))=0.207"
            )

        self.check_params(parameters)

        # Check if user requests derivative splines
        if self.check_param_derivatives(parameter_sensitivities):
            n_derivatives = parameter_sensitivities.shape[2]
            derivatives = []
        else:
            n_derivatives = 0
            derivatives = None

        splines = []
        for i_derivative in range(n_derivatives + 1):
            # Constant auxiliary values
            if i_derivative == 0:
                cl = contact_length
                thick_vert_hor = parameters[0, 0]  # parameters.shape == [1]
                thick_diagonal = parameters[0, 1]
                v_one_half = 0.5
                v_one = 1.0
                v_zero = 0.0
            else:
                cl = 0.0
                thick_vert_hor = parameter_sensitivities[
                    0, 0, i_derivative - 1
                ]
                thick_diagonal = parameter_sensitivities[
                    0, 1, i_derivative - 1
                ]
                v_one_half = 0.0
                v_one = 0.0
                v_zero = 0.0

            # Set variables
            a01 = v_zero
            a02 = thick_vert_hor
            a03 = thick_vert_hor + thick_diagonal * _np.sqrt(2)
            a04 = (v_one - cl) * 0.5
            a05 = v_one_half - thick_diagonal * _np.sqrt(2)
            a06 = v_one_half
            a07 = v_one_half + thick_diagonal * _np.sqrt(2)
            a08 = (v_one + cl) * 0.5
            a09 = v_one - (thick_vert_hor + thick_diagonal * _np.sqrt(2))
            a10 = v_one - thick_vert_hor
            a11 = v_one
            # Init return value
            spline_list = []

            # 1
            spline_list.append(
                _Bezier(
                    degrees=[1, 1],
                    control_points=[
                        [a01, a01],
                        [a02, a02],
                        [a01, a04],
                        [a02, a03],
                    ],
                )
            )

            # 2
            spline_list.append(
                _Bezier(
                    degrees=[1, 1],
                    control_points=[
                        [a01, a01],
                        [a04, a01],
                        [a02, a02],
                        [a03, a02],
                    ],
                )
            )

            # 3
            spline_list.append(
                _Bezier(
                    degrees=[1, 1],
                    control_points=[
                        [a04, a01],
                        [a08, a01],
                        [a03, a02],
                        [a09, a02],
                    ],
                )
            )

            # 4
            spline_list.append(
                _Bezier(
                    degrees=[1, 1],
                    control_points=[
                        [a08, a01],
                        [a11, a01],
                        [a09, a02],
                        [a10, a02],
                    ],
                ),
            )

            # 5
            spline_list.append(
                _Bezier(
                    degrees=[1, 1],
                    control_points=[
                        [a10, a02],
                        [a11, a01],
                        [a10, a03],
                        [a11, a04],
                    ],
                )
            )

            # 6
            spline_list.append(
                _Bezier(
                    degrees=[1, 1],
                    control_points=[
                        [a11, a04],
                        [a11, a08],
                        [a10, a03],
                        [a10, a09],
                    ],
                )
            )

            # 7
            spline_list.append(
                _Bezier(
                    degrees=[1, 1],
                    control_points=[
                        [a10, a09],
                        [a11, a08],
                        [a10, a10],
                        [a11, a11],
                    ],
                )
            )
            # 8
            spline_list.append(
                _Bezier(
                    degrees=[1, 1],
                    control_points=[
                        [a09, a10],
                        [a10, a10],
                        [a08, a11],
                        [a11, a11],
                    ],
                )
            )

            # 9
            spline_list.append(
                _Bezier(
                    degrees=[1, 1],
                    control_points=[
                        [a03, a10],
                        [a09, a10],
                        [a04, a11],
                        [a08, a11],
                    ],
                )
            )

            # 10
            spline_list.append(
                _Bezier(
                    degrees=[1, 1],
                    control_points=[
                        [a02, a10],
                        [a03, a10],
                        [a01, a11],
                        [a04, a11],
                    ],
                )
            )

            # 11
            spline_list.append(
                _Bezier(
                    degrees=[1, 1],
                    control_points=[
                        [a01, a08],
                        [a02, a09],
                        [a01, a11],
                        [a02, a10],
                    ],
                )
            )

            # 12
            spline_list.append(
                _Bezier(
                    degrees=[1, 1],
                    control_points=[
                        [a01, a04],
                        [a02, a03],
                        [a01, a08],
                        [a02, a09],
                    ],
                )
            )

            # 13
            spline_list.append(
                _Bezier(
                    degrees=[1, 1],
                    control_points=[
                        [a02, a09],
                        [a05, a06],
                        [a02, a10],
                        [a06, a06],
                    ],
                )
            )

            # 14
            spline_list.append(
                _Bezier(
                    degrees=[1, 1],
                    control_points=[
                        [a02, a10],
                        [a06, a06],
                        [a03, a10],
                        [a06, a07],
                    ],
                )
            )

            # 15
            spline_list.append(
                _Bezier(
                    degrees=[1, 1],
                    control_points=[
                        [a02, a02],
                        [a06, a06],
                        [a02, a03],
                        [a05, a06],
                    ],
                )
            )

            # 16
            spline_list.append(
                _Bezier(
                    degrees=[1, 1],
                    control_points=[
                        [a02, a02],
                        [a03, a02],
                        [a06, a06],
                        [a06, a05],
                    ],
                )
            )

            # 17
            spline_list.append(
                _Bezier(
                    degrees=[1, 1],
                    control_points=[
                        [a09, a02],
                        [a10, a02],
                        [a06, a05],
                        [a06, a06],
                    ],
                )
            )

            # 18
            spline_list.append(
                _Bezier(
                    degrees=[1, 1],
                    control_points=[
                        [a06, a06],
                        [a10, a02],
                        [a07, a06],
                        [a10, a03],
                    ],
                )
            )

            # 19
            spline_list.append(
                _Bezier(
                    degrees=[1, 1],
                    control_points=[
                        [a06, a06],
                        [a07, a06],
                        [a10, a10],
                        [a10, a09],
                    ],
                )
            )

            # 20
            spline_list.append(
                _Bezier(
                    degrees=[1, 1],
                    control_points=[
                        [a06, a06],
                        [a10, a10],
                        [a06, a07],
                        [a09, a10],
                    ],
                )
            )

            # Pass to output
            if i_derivative == 0:
                splines = spline_list.copy()
            else:
                derivatives.append(spline_list)

        # Return results
        return (splines, derivatives)



import base64 as _base64
import copy as _copy
import xml.etree.ElementTree as _ET
from sys import version as _python_version

import numpy as _np

from splinepy import settings as _settings
from splinepy.utils.data import enforce_contiguous as _enforce_contiguous
from splinepy.utils.log import debug as _debug
from splinepy.utils.log import warning as _warning


def _spline_to_ET(
    root,
    multipatch,
    index_offset,
    fields_only=False,
    as_base64=False,
    field_mask=None,
):
    """
    Write spline data to xml element in gismo format

    Parameters
    ----------
    root : ElementTree.Subelement
      branch in element tree to which the spline info is to be added
    multipatch : Multipatch
      Multipatch containing the information requested
    index_offset : int
      index_offset for ids in xml export
      (All geometries are assigned individual ids that must be unique in the
      xml file)
    fields_only : bool (False)
      If set, exports only the fields associated to the multipatch data
    as_base64 : bool (False)
      If set, coordinates, weights and knot-vectors are exported in B64
    field_mask : array-like
      Selection of active fields to be exported into the fields file

    Returns
    -------
    None
    """
    from splinepy.spline import Spline as _Spline

    if fields_only and (len(multipatch.fields) == 0):
        return

    if fields_only and field_mask is not None:
        if not isinstance(field_mask, (list, _np.ndarray)):
            raise ValueError("field_mask must be list of integers")
        field_mask = _np.unique(
            _enforce_contiguous(field_mask, dtype=_np.int64, asarray=True)
        )

        # Check if range is valid
        if (field_mask.min() < 0) or (
            field_mask.max() >= len(multipatch.fields)
        ):
            raise ValueError(
                "field_mask contains unsupported range, must be within "
                f"(0, {len(multipatch.fields)})"
            )

    def _array_to_text(array, is_matrix, as_b64):
        if as_b64:
            array = _enforce_contiguous(array, dtype=_np.float64, asarray=True)
            return _base64.b64encode(array).decode("ascii")
        elif is_matrix:
            return "\n".join(
                [" ".join([str(x) for x in row]) for row in array]
            )
        else:
            return " ".join(str(x) for x in array)

    if fields_only:
        if field_mask is None:
            field_mask = _np.arange(len(multipatch.fields))
        supports = _np.array(
            [
                (i, j, 0)
                for j, k in enumerate(field_mask)
                for i, v in enumerate(multipatch.fields[k].patches)
                if v is not None
            ],
            dtype=_np.int64,
        )

        # Sanity check (see error message)
        n_patches_in_support = _np.max(supports[:, 0]) + 1
        if _np.unique(supports[:, 0]).size != n_patches_in_support:
            raise ValueError(
                "The requested field(s) can not be exported, because the patch "
                "ids of the field support (i.e., where the fields are non-zero)"
                " is not contiguous. This can be the case, if a field is only "
                "non-zero on patches [0,1,4]. This case is currently not "
                "supported. If required, it is possible to (1) change the input"
                " routine and export a patch-mapping (former patch id to new "
                "patch id or (2) add zero-valued splines to fill the gaps. "
                "Please feel free to write an issue on github.com/taratat/"
                "splinepy for additional information."
            )

        # Very unintuitive solution to counting the support ids #indextrick
        indices = _np.argsort(supports[:, 0], kind="stable")
        counter = _np.arange(supports.shape[0])
        bincount = _np.bincount(supports[:, 0])
        # Get number of occurrences for previous element to correct index shift
        index_shift = _np.cumsum(_np.hstack([[0], bincount[:-1]]))
        counter -= _np.repeat(index_shift, bincount)
        supports[indices, 2] = counter

        # Write Matrix
        design_v_support = _ET.SubElement(
            root,
            "Matrix",
            cols=str(supports.shape[1]),
            id=str(10),
            rows=str(supports.shape[0]),
        )
        design_v_support.text = "\n".join(
            [" ".join([str(xx) for xx in x]) for x in supports]
        )
    else:
        n_patches_in_support = len(multipatch.patches)

    for id, spline in enumerate(multipatch.patches):
        if fields_only:
            # Check supports
            support = field_mask[supports[supports[:, 0] == id, 1]]
            if support.size == 0:
                continue
            coefs = _np.hstack(
                [
                    multipatch.fields[j].patches[id].control_points
                    for j in support
                ]
            )
            if "weights" in spline.required_properties:
                weights = _np.hstack(
                    [multipatch.fields[j][id].weights for j in support]
                )
        else:
            coefs = spline.control_points
            if "weights" in spline.required_properties:
                weights = spline.weights

        if not issubclass(type(spline), _Spline):
            raise ValueError(
                "One of the splines handed to export is not a valid spline"
                " representation"
            )

        # Transform bezier types, as they are not supported in gismo
        if spline.name.startswith("Bezier"):
            type_name = "BSpline"
            spline = spline.bspline  # noqa PLW2901
        elif spline.name.startswith("RationalBezier"):
            type_name = "Nurbs"
            spline = spline.nurbs  # noqa PLW2901
        elif spline.name.startswith("BSpline"):
            type_name = "BSpline"
        elif spline.name.startswith("NURBS"):
            type_name = "Nurbs"

        # Start element definition
        spline_element = _ET.SubElement(
            root,
            "Geometry",
            id=str(id + index_offset),
            type="Tensor" + type_name + str(spline.para_dim),
        )

        # Define Basis functions
        if "weights" in spline.required_properties:
            spline_basis_base = _ET.SubElement(
                spline_element,
                "Basis",
                type="Tensor" + type_name + "Basis" + str(spline.para_dim),
            )
            spline_basis = _ET.SubElement(
                spline_basis_base,
                "Basis",
                type="TensorBSplineBasis" + str(spline.para_dim),
            )
        else:
            spline_basis = _ET.SubElement(
                spline_element,
                "Basis",
                type="Tensor" + type_name + "Basis" + str(spline.para_dim),
            )

        # Set the format flag for output type
        format_flag = "B64Float64" if as_base64 else "ASCII"

        for i_para in range(spline.para_dim):
            basis_fun = _ET.SubElement(
                spline_basis,
                "Basis",
                index=str(i_para),
                type="BSplineBasis",
            )
            knot_vector = _ET.SubElement(
                basis_fun,
                "KnotVector",
                degree=str(spline.degrees[i_para]),
                format=format_flag,
            )
            knot_vector.text = _array_to_text(
                spline.knot_vectors[i_para], False, as_base64
            )

        if "weights" in spline.required_properties:
            # Add weights
            weights_element = _ET.SubElement(
                spline_basis_base,
                "weights",
                format=format_flag,
            )
            weights_element.text = _array_to_text(weights, True, as_base64)
        coords = _ET.SubElement(
            spline_element,
            "coefs",
            format=format_flag,
            geoDim=str(coefs.shape[1]),
        )
        coords.text = _array_to_text(coefs, True, as_base64)

    return n_patches_in_support


def export(
    fname,
    multipatch=None,
    indent=True,
    labeled_boundaries=True,
    options=None,
    export_fields=False,
    field_mask=None,
    as_base64=False,
):
    """Export as gismo readable xml geometry file
    Use gismo-specific xml-keywords to export (list of) splines. All Bezier
    patches are excluded as their respective non uniform counterpart

    Parameters
    ----------
    fname : string
      name of the output file
    spline_list : Multipatch (preferred)
      (list of) Spline-Types in splinepy format
    indent: bool
      Appends white spaces using xml.etree.ElementTree.indent, if possible.
    labeled_boundaries : bool
      Writes boundaries with labels into the MultiPatch part of the XML
    options : list
      List of dictionaries, that specify model related options, like boundary
      conditions, assembler options, etc.. The dictionaries must have the
      following keys, 'tag'->string, 'text'->string (optional),
      'attributes'->dictionary (optional), 'children'->list in the same format
      (optional)
    export_fields : bool
      Export fields to separate file ending with fields.xml, e.g.,
      filename.xml.fields.xml. Only non-zero splines are exported to save
      memory
    field_mask : list
      Selection of active fields that are exported (to save memory and speed up
      export)
    as_base64 : bool (False)
      If set, coordinates, weights and knot-vectors are exported in B64
      encoding. Available in gismo, once gismo/gismo#634 merges.

    Returns
    -------
    None
    """
    from splinepy import Multipatch as _Multipatch
    from splinepy.spline import Spline as _Spline

    # First transform spline-data into a multipatch-data if required
    if issubclass(type(multipatch), _Spline):
        # Transform to list
        multipatch = [multipatch]

    if isinstance(multipatch, list):
        # Transform to multipatch
        multipatch = _Multipatch(splines=multipatch)
        _warning(
            "No interfaces (i.e. inter-face connectivity) information is"
            " given. Try converting list to Multipatch object. Interfaces"
            "will be computed on the fly as a result."
        )

    if not isinstance(multipatch, _Multipatch):
        raise ValueError(
            "export Function expects list for multipatch argument"
        )

    if not isinstance(fname, str):
        raise ValueError("fname argument must be string")

    # All entries in gismo xml file have a unique ID
    xml_data = _ET.Element("xml")
    index_offset = 100

    # licked it so its ours
    xml_data.insert(
        1,
        _ET.Comment(
            "generated by splinepy https://github.com/tataratat/splinepy"
        ),
    )

    # First export Multipatch information
    multipatch_element = _ET.SubElement(
        xml_data, "MultiPatch", id=str(0), parDim=str(multipatch.para_dim)
    )
    patch_range = _ET.SubElement(
        multipatch_element, "patches", type="id_range"
    )

    # Retrieve all interfaces (negative numbers refer to boundaries)
    interface_data = _ET.SubElement(multipatch_element, "interfaces")
    global_interface_id = _np.where(multipatch.interfaces.ravel() >= 0)[0]

    if global_interface_id.size == 0:
        _warning("No inter-face connections were found.")
    else:
        # Reminder: Face enumeration starts at 1 in gismo (i.e. requires an
        # offset of 1) Copy to not overwrite original interface orientations
        interface_array = multipatch.interface_orientations().copy()

        # Apply offsets
        interface_array[:, [0, 2]] += index_offset
        interface_array[:, [1, 3]] += 1  # Gismo starts counting at 1

        interface_data.text = "\n".join(
            [" ".join([str(xx) for xx in x]) for x in interface_array]
        )

    if labeled_boundaries:
        # Export geometric boundaries and set a label
        for boundary_id, face_id_list in enumerate(
            multipatch.boundaries, start=1
        ):
            boundary_data = _ET.SubElement(
                multipatch_element,
                "boundary",
                name="BID" + str(boundary_id),
            )
            boundary_data.text = "\n".join(
                [
                    str(patch_id + index_offset) + " " + str(local_face_id + 1)
                    for (patch_id, local_face_id) in zip(*face_id_list)
                ]
            )
    else:
        # This is valid in the old format (1. list all boundaries, 2. set BCs)
        # Start extracting all boundaries
        # Boundaries are stored as patch-id (global) face-id (local)
        boundary_spline, boundary_face = _np.where(multipatch.interfaces < 0)
        if boundary_spline.size > 0:
            boundary_data = _ET.SubElement(multipatch_element, "boundary")
            boundary_data.text = "\n".join(
                [
                    str(sid + index_offset) + " " + str(bid + 1)
                    for (sid, bid) in zip(boundary_spline, boundary_face)
                ]
            )
        ###
        # Boundary conditions
        ###
        # If there are multiple boundary IDs write them into boundary
        # conditions and fill with dummy values
        # Note here, that the boundary conditions are always stored in the
        # local enumeration system of them multipatch
        boundary_condition_list = multipatch.boundaries
        if len(boundary_condition_list) > 1:
            bcs_data = _ET.SubElement(
                xml_data,
                "boundaryConditions",
                id=str(1),
                multipatch=str(len(multipatch.patches)),
            )
            bcs_data.insert(
                0, _ET.Comment(text="Please fill boundary conditions here")
            )
            for bc_data_i in boundary_condition_list:
                bc = _ET.SubElement(
                    bcs_data, "bc", type="Dirichlet", unknown="0"
                )
                bc.text = "\n".join(
                    [
                        str(sid) + " " + str(bid + 1)
                        for (sid, bid) in zip(bc_data_i[0], bc_data_i[1])
                    ]
                )

    ###
    # Individual spline data
    ###
    # Export fields first, as all necessary information is already available
    if export_fields:
        field_xml = _copy.deepcopy(xml_data)

        n_patches = _spline_to_ET(
            field_xml,
            multipatch,
            index_offset,
            fields_only=True,
            as_base64=as_base64,
            field_mask=field_mask,
        )
        field_xml.find("MultiPatch").find("patches").text = (
            f"{index_offset} " f"{n_patches - 1  + index_offset}"
        )
        if int(_python_version.split(".")[1]) >= 9 and indent:
            _ET.indent(field_xml)
        file_content = _ET.tostring(field_xml)
        with open(fname + ".fields.xml", "wb") as f:
            f.write(file_content)

    n_patches = _spline_to_ET(
        xml_data,
        multipatch,
        index_offset,
        as_base64=as_base64,
    )

    if n_patches != len(multipatch.patches):
        raise RuntimeError("Help - some patches were not recognised")

    patch_range.text = f"{index_offset} " f"{n_patches - 1 + index_offset}"

    # Add additional options to the xml file
    if options is not None:
        # Verify that the list stored in the correct format
        if not isinstance(options, list):
            options = [options]

        def _apply_options(ETelement, options_list):
            for gismo_dictionary in options_list:
                name = gismo_dictionary.get("tag", None)
                if name is None and not isinstance(name, str):
                    raise ValueError(
                        "Gismo option in unsupported format, tag must be set, "
                        "please check out export documentation"
                    )
                attributes = gismo_dictionary.get("attributes", {})
                option_text = gismo_dictionary.get("text", None)
                optional_data = _ET.SubElement(
                    ETelement,
                    name,
                    **attributes,
                )
                if option_text is not None:
                    optional_data.text = option_text
                if gismo_dictionary.get("children", None) is not None:
                    _apply_options(
                        optional_data,
                        options_list=gismo_dictionary["children"],
                    )

        _apply_options(xml_data, options)

    if int(_python_version.split(".")[1]) >= 9 and indent:
        # Pretty printing xml with indent only exists in version > 3.9
        _ET.indent(xml_data)

    elif int(_python_version.split(".")[1]) < 9 and indent:
        _debug(
            "Indented xml output is only supported from > python3.9.",
            "Output will not be indented.",
            f"Current python version: {_python_version}",
        )

    file_content = _ET.tostring(xml_data)
    with open(fname, "wb") as f:
        f.write(file_content)


def load(fname, load_options=True):
    """Read gismo-keyword specified xml file

    Parameters
    ----------
    fname : str
      filename of the gismo xml
    load_options : bool
      Retrieve additional options (else - multipatch/geometry only)

    Returns
    -------
    spline_dic_list : Multipatch
      Multipatch object with list of splines in NAME_TO_TYPE-type
    options : list
      List of additional options, like boundary conditions and functions
    """
    from splinepy.multipatch import Multipatch as _Multipatch

    def _matrix_from_node(xml_node):
        """
        Parameters
        ----------
        xml_node : XMLElementTree

        Returns
        -------
        array : np.array
          numpy array in unspecified type, one-dimensional
        """
        type_from_keyword = {
            "b64float64": _np.float64,
            "b64float32": _np.float32,
            "b64uint16": _np.uint16,
            "b64uint32": _np.uint32,
            "b64uint64": _np.uint64,
            "b64int16": _np.int16,
            "b64int32": _np.int32,
            "b64int64": _np.int64,
        }

        # Check format
        format_flag = xml_node.attrib.get("format", "ASCII")
        if format_flag.lower() == "ascii":
            return _np.fromstring(xml_node.text.replace("\n", " "), sep=" ")
        # Check for all excepted binary flags
        format = type_from_keyword.get(format_flag.lower(), None)

        if format is None:
            raise ValueError(
                "Unknown format in xml file: " + format_flag.lower()
            )

        return _np.frombuffer(
            _base64.b64decode(xml_node.text.strip().encode("ascii")),
            dtype=_np.float64,
        )

    # Auxiliary function to unravel
    def retrieve_from_basis_(ETelement, SPdict):
        if "nurbs" in ETelement.attrib["type"].lower():
            # Recursive call for knot_vector
            for child in ETelement:
                if "weights" in child.tag:
                    SPdict["weights"] = _matrix_from_node(child)
                if "basis" in child.tag.lower():
                    retrieve_from_basis_(child, SPdict)
        elif "bspline" in ETelement.attrib["type"].lower():
            para_dim = int(ETelement.attrib["type"].lower().split("basis")[-1])
            knotvector = [None] * para_dim
            degrees = [None] * para_dim
            for child in ETelement:
                if "bsplinebasis" not in child.attrib["type"].lower():
                    raise ValueError("Something went wrong")
                if "knotvector" in child[0].tag.lower():
                    id = int(child.attrib["index"])
                    kv = _matrix_from_node(child[0])
                    knotvector[id] = kv
                    degrees[id] = int(child[0].attrib["degree"])
            SPdict["knot_vectors"] = knotvector
            SPdict["degrees"] = degrees

    # Auxiliary function to store additional information in dictionary
    def make_dictionary(ETelement):
        new_dictionary = {}
        new_dictionary["tag"] = ETelement.tag
        new_dictionary["attributes"] = ETelement.attrib
        new_dictionary["text"] = ETelement.text
        new_dictionary["children"] = []
        for element in ETelement.findall("./*"):
            new_dictionary["children"].append(make_dictionary(element))
        return new_dictionary

    # Parse XML file
    _debug(f"Parsing xml-file '{fname}' ...")
    root = _ET.parse(fname).getroot()
    _debug("XML-file parsed start conversion")

    # Init return value
    list_of_splines = []
    list_of_options = []
    interface_array = None
    invalid_integer = -99919412
    # Splines start with the keyword Geometry
    for child in root:
        if child.tag.startswith("MultiPatch"):
            parametric_dimension = int(child.attrib["parDim"])
            # Patch ids
            patch_element = child.find("patches")
            if patch_element is None:
                _debug("Unsupported format")
            if patch_element.attrib.get("type") != "id_range":
                _debug(
                    f"Invalid patch type {patch_element.attrib.get('type')}"
                )
            patch_range = _matrix_from_node(patch_element).astype(_np.int64)
            offset = patch_range[0]
            n_splines = patch_range[1] - patch_range[0] + 1
            number_of_faces = parametric_dimension * 2

            # Extract interfaces and boundaries
            interfaces_element = child.find("interfaces")
            interface_array = (
                _np.ones((n_splines, number_of_faces), dtype=_np.int64)
                * invalid_integer
            )
            if interfaces_element is None or interfaces_element.text is None:
                # This can occur if only one patch is in a Multipatch
                _debug("No connectivity found")
            else:
                interface_information = (
                    _matrix_from_node(interfaces_element)
                    .astype(
                        dtype=_np.int64,
                    )
                    .reshape(-1, number_of_faces + 4)
                )
                # Assign interfaces
                interface_array[
                    interface_information[:, 0] - offset,
                    interface_information[:, 1] - 1,
                ] = (
                    (interface_information[:, 2] - offset) * number_of_faces
                    + interface_information[:, 3]
                    - 1
                )
                interface_array[
                    interface_information[:, 2] - offset,
                    interface_information[:, 3] - 1,
                ] = (
                    (interface_information[:, 0] - offset) * number_of_faces
                    + interface_information[:, 1]
                    - 1
                )

            # Extract interfaces and boundaries
            boundary_elements = child.findall("boundary")
            if boundary_elements is None:
                _debug("No boundary information found")
            else:
                for id, boundary_element in enumerate(
                    boundary_elements, start=1
                ):
                    if boundary_element.text is None:
                        continue
                    boundary_information = (
                        _matrix_from_node(boundary_element)
                        .astype(
                            dtype=_np.int64,
                        )
                        .reshape(-1, 2)
                    )
                    interface_array[
                        boundary_information[:, 0] - offset,
                        boundary_information[:, 1] - 1,
                    ] = -id

        elif child.tag.startswith("Geometry"):
            spline_dict = {}
            _debug(f"Found new spline in xml with id {child.attrib.get('id')}")
            for info in child:
                if "basis" in info.tag.lower():
                    retrieve_from_basis_(info, spline_dict)
                elif info.tag.startswith("coef"):
                    dim = int(info.attrib.get("geoDim"))
                    spline_dict["control_points"] = _matrix_from_node(
                        info
                    ).reshape(-1, dim)
            if spline_dict.get("weights") is None:
                list_of_splines.append(
                    _settings.NAME_TO_TYPE["BSpline"](**spline_dict)
                )
            else:
                list_of_splines.append(
                    _settings.NAME_TO_TYPE["NURBS"](**spline_dict)
                )
        elif load_options:
            list_of_options.append(make_dictionary(child))
        else:
            _debug(
                f"Found unsupported keyword {child.tag}, which will be"
                " ignored"
            )
            continue

    _debug(f"Found a total of {len(list_of_splines)} " f"BSplines and NURBS")
    multipatch = _Multipatch(list_of_splines)
    if interface_array is not None:
        if invalid_integer in interface_array:
            _warning("Unmatched faces or insufficient information in xml file")
            invalid_splines = _np.where(interface_array)[0][0] + offset
            _warning(f"Check faces with ids {invalid_splines}")
            _warning("Interface information ignored")

        multipatch.interfaces = interface_array

    if load_options:
        return multipatch, list_of_options
    else:
        return multipatch



import numpy as _np

from splinepy.bezier import Bezier as _Bezier
from splinepy.microstructure.tiles.tile_base import TileBase as _TileBase


class HollowOctagonExtrude(_TileBase):
    """Simple tile - looks like a nut, but in 3D.

    .. raw:: html

        <p><a href="../_static/HollowOctagonExtrude.html">Fullscreen</a>.</p>
        <embed type="text/html" width="100%" height="400" src="../_static/HollowOctagonExtrude.html" />

    """  # noqa: E501

    _dim = 3
    _para_dim = 3
    _evaluation_points = _np.array([[0.5, 0.5, 0.5]])
    _n_info_per_eval_point = 1

    def create_tile(
        self,
        parameters=None,
        parameter_sensitivities=None,  # TODO
        contact_length=0.2,
        **kwargs,  # noqa ARG002
    ):
        """Create a microtile based on the parameters that describe the wall
        thicknesses.

        Thickness parameters are used to describe the inner radius of the
        outward facing branches

        Parameters
        ----------
        parameters : np.array(1, 1)
          One evaluation point with one parameter is used. This parameter
          specifies the distance from the center to the inner edge, where
          the value must be between 0.01 and 0.49.
        parameter_sensitivities: np.ndarray
          Describes the parameter sensitivities with respect to some design
          variable. In case the design variables directly apply to the
          parameter itself, they evaluate as delta_ij
        contact_length : float
            the length of the wall that contacts the other microstructure
        Returns
        -------
        microtile_list : list(splines)
        derivatives: list<list<splines>> / None
        """

        if not isinstance(contact_length, float):
            raise ValueError("Invalid Type for radius")

        if not ((contact_length > 0) and (contact_length < 0.99)):
            raise ValueError("The length of a side must be in (0.01, 0.99)")

        if parameters is None:
            self._logd("Setting parameters to default values (0.2)")
            parameters = _np.array(
                _np.ones(
                    (len(self._evaluation_points), self._n_info_per_eval_point)
                )
                * 0.2
            )

        if parameter_sensitivities is not None:
            raise NotImplementedError(
                "Derivatives are not implemented for this tile yet"
            )

        self.check_params(parameters)

        v_h_void = parameters[0, 0]
        if not ((v_h_void > 0.01) and (v_h_void < 0.5)):
            raise ValueError(
                "The thickness of the wall must be in (0.01 and 0.49)"
            )

        v_zero = 0.0
        v_one_half = 0.5
        v_one = 1.0
        v_outer_c_h = contact_length * 0.5
        v_inner_c_h = contact_length * parameters[0, 0]

        spline_list = []

        # set points:
        right = _np.array(
            [
                [v_h_void + v_one_half, -v_inner_c_h + v_one_half, v_zero],
                [v_one, -v_outer_c_h + v_one_half, 0.0],
                [v_h_void + v_one_half, v_inner_c_h + v_one_half, 0.0],
                [v_one, v_outer_c_h + v_one_half, 0.0],
                [v_h_void + v_one_half, -v_inner_c_h + v_one_half, v_one],
                [v_one, -v_outer_c_h + v_one_half, v_one],
                [v_h_void + v_one_half, v_inner_c_h + v_one_half, v_one],
                [v_one, v_outer_c_h + v_one_half, v_one],
            ]
        )

        right_top = _np.array(
            [
                [v_h_void + v_one_half, v_inner_c_h + v_one_half, v_zero],
                [v_one, v_outer_c_h + v_one_half, v_zero],
                [v_inner_c_h + v_one_half, v_h_void + v_one_half, v_zero],
                [v_outer_c_h + v_one_half, v_one, v_zero],
                [v_h_void + v_one_half, v_inner_c_h + v_one_half, v_one],
                [v_one, v_outer_c_h + v_one_half, v_one],
                [v_inner_c_h + v_one_half, v_h_void + v_one_half, v_one],
                [v_outer_c_h + v_one_half, v_one, v_one],
            ]
        )

        top = _np.array(
            [
                [v_inner_c_h + v_one_half, v_h_void + v_one_half, v_zero],
                [v_outer_c_h + v_one_half, v_one, v_zero],
                [-v_inner_c_h + v_one_half, v_h_void + v_one_half, v_zero],
                [-v_outer_c_h + v_one_half, v_one, v_zero],
                [v_inner_c_h + v_one_half, v_h_void + v_one_half, v_one],
                [v_outer_c_h + v_one_half, v_one, v_one],
                [-v_inner_c_h + v_one_half, v_h_void + v_one_half, v_one],
                [-v_outer_c_h + v_one_half, v_one, v_one],
            ]
        )

        bottom_left = _np.array(
            [
                [-v_h_void + v_one_half, -v_inner_c_h + v_one_half, v_zero],
                [v_zero, -v_outer_c_h + v_one_half, v_zero],
                [-v_inner_c_h + v_one_half, -v_h_void + v_one_half, v_zero],
                [-v_outer_c_h + v_one_half, v_zero, v_zero],
                [-v_h_void + v_one_half, -v_inner_c_h + v_one_half, v_one],
                [v_zero, -v_outer_c_h + v_one_half, v_one],
                [-v_inner_c_h + v_one_half, -v_h_void + v_one_half, v_one],
                [-v_outer_c_h + v_one_half, v_zero, v_one],
            ]
        )

        left = _np.array(
            [
                [v_zero, -v_outer_c_h + v_one_half, v_zero],
                [-v_h_void + v_one_half, -v_inner_c_h + v_one_half, v_zero],
                [v_zero, v_outer_c_h + v_one_half, v_zero],
                [-v_h_void + v_one_half, v_inner_c_h + v_one_half, v_zero],
                [v_zero, -v_outer_c_h + v_one_half, v_one],
                [-v_h_void + v_one_half, -v_inner_c_h + v_one_half, v_one],
                [v_zero, v_outer_c_h + v_one_half, v_one],
                [-v_h_void + v_one_half, v_inner_c_h + v_one_half, v_one],
            ]
        )

        top_left = _np.array(
            [
                [v_zero, v_outer_c_h + v_one_half, v_zero],
                [-v_h_void + v_one_half, v_inner_c_h + v_one_half, v_zero],
                [-v_outer_c_h + v_one_half, v_one, v_zero],
                [-v_inner_c_h + v_one_half, v_h_void + v_one_half, v_zero],
                [v_zero, v_outer_c_h + v_one_half, v_one],
                [-v_h_void + v_one_half, v_inner_c_h + v_one_half, v_one],
                [-v_outer_c_h + v_one_half, v_one, v_one],
                [-v_inner_c_h + v_one_half, v_h_void + v_one_half, v_one],
            ]
        )

        bottom = _np.array(
            [
                [v_outer_c_h + v_one_half, v_zero, v_zero],
                [v_inner_c_h + v_one_half, -v_h_void + v_one_half, v_zero],
                [-v_outer_c_h + v_one_half, v_zero, v_zero],
                [-v_inner_c_h + v_one_half, -v_h_void + v_one_half, v_zero],
                [v_outer_c_h + v_one_half, v_zero, v_one],
                [v_inner_c_h + v_one_half, -v_h_void + v_one_half, v_one],
                [-v_outer_c_h + v_one_half, v_zero, v_one],
                [-v_inner_c_h + v_one_half, -v_h_void + v_one_half, v_one],
            ]
        )

        bottom_right = _np.array(
            [
                [v_inner_c_h + v_one_half, -v_h_void + v_one_half, v_zero],
                [v_outer_c_h + v_one_half, v_zero, v_zero],
                [v_h_void + v_one_half, -v_inner_c_h + v_one_half, v_zero],
                [v_one, -v_outer_c_h + v_one_half, v_zero],
                [v_inner_c_h + v_one_half, -v_h_void + v_one_half, v_one],
                [v_outer_c_h + v_one_half, v_zero, v_one],
                [v_h_void + v_one_half, -v_inner_c_h + v_one_half, v_one],
                [v_one, -v_outer_c_h + v_one_half, v_one],
            ]
        )

        spline_list.append(_Bezier(degrees=[1, 1, 1], control_points=right))

        spline_list.append(
            _Bezier(degrees=[1, 1, 1], control_points=right_top)
        )

        spline_list.append(_Bezier(degrees=[1, 1, 1], control_points=bottom))

        spline_list.append(
            _Bezier(degrees=[1, 1, 1], control_points=bottom_left)
        )

        spline_list.append(_Bezier(degrees=[1, 1, 1], control_points=left))

        spline_list.append(_Bezier(degrees=[1, 1, 1], control_points=top_left))

        spline_list.append(_Bezier(degrees=[1, 1, 1], control_points=top))

        spline_list.append(
            _Bezier(degrees=[1, 1, 1], control_points=bottom_right)
        )

        return (spline_list, None)

    def closing_tile(
        self,
        parameters=None,
        parameter_sensitivities=None,  # TODO
        contact_length=0.2,
        closure=None,
    ):
        """Create a closing tile to match with closed surface.

        Parameters
        ----------
        parameters: np.ndarray(1, 1)
          One evaluation point with one parameter is used. This parameter
          specifies the distance from the center to the inner edge, where
          the value must be between 0.01 and 0.49.
        parameter_sensitivities: np.ndarray
          Describes the parameter sensitivities with respect to some design
          variable. In case the design variables directly apply to the
          parameter itself, they evaluate as delta_ij
        closure : int
          parametric dimension that needs to be closed. Positive values mean
          that minimum parametric dimension is requested. That means,
          i.e. -2 closes the tile at maximum z-coordinate.
          (must currently be either -2 or 2)
        contact_length: float
          the length of the wall that contacts the other microstructure

        Returns
        -------
        spline_list : list
        derivatives: list<list<splines>> / None
        """
        if closure is None:
            raise ValueError("No closing direction given")

        if parameters is None:
            self._log("Tile request is not parametrized, setting default 0.2")
            parameters = _np.array(
                _np.ones(
                    (len(self._evaluation_points), self._n_info_per_eval_point)
                )
                * 0.2
            )

        if not (_np.all(parameters[0] > 0) and _np.all(parameters[0] < 0.5)):
            raise ValueError(
                "The thickness of the wall must be in (0.01 and 0.49)"
            )

        self.check_params(parameters)

        if parameter_sensitivities is not None:
            raise NotImplementedError(
                "Derivatives are not implemented for this tile yet"
            )

        v_h_void = parameters[0, 0]
        if not ((v_h_void > 0.01) and (v_h_void < 0.5)):
            raise ValueError(
                "The thickness of the wall must be in (0.01 and 0.49)"
            )

        spline_list = []
        v_zero = 0.0
        v_one_half = 0.5
        v_one = 1.0
        v_outer_c_h = contact_length * 0.5
        v_inner_c_h = contact_length * parameters[0, 0]

        if closure == "x_min":
            # set points:
            right = _np.array(
                [
                    [v_h_void + v_one_half, -v_inner_c_h + v_one_half],
                    [v_one, -v_outer_c_h + v_one_half],
                    [v_h_void + v_one_half, v_inner_c_h + v_one_half],
                    [v_one, v_outer_c_h + v_one_half],
                ]
            )

            right_top = _np.array(
                [
                    [v_h_void + v_one_half, v_inner_c_h + v_one_half],
                    [v_one, v_outer_c_h + v_one_half],
                    [v_inner_c_h + v_one_half, v_h_void + v_one_half],
                    [v_outer_c_h + v_one_half, v_one],
                ]
            )

            top = _np.array(
                [
                    [v_inner_c_h + v_one_half, v_h_void + v_one_half],
                    [v_outer_c_h + v_one_half, v_one],
                    [-v_inner_c_h + v_one_half, v_h_void + v_one_half],
                    [-v_outer_c_h + v_one_half, v_one],
                ]
            )

            bottom_left = _np.array(
                [
                    [-v_h_void + v_one_half, -v_inner_c_h + v_one_half],
                    [v_zero, v_zero],
                    [-v_inner_c_h + v_one_half, -v_h_void + v_one_half],
                    [-v_outer_c_h + v_one_half, v_zero],
                ]
            )

            left = _np.array(
                [
                    [v_zero, v_zero],
                    [-v_h_void + v_one_half, -v_inner_c_h + v_one_half],
                    [v_zero, v_one],
                    [-v_h_void + v_one_half, v_inner_c_h + v_one_half],
                ]
            )

            top_left = _np.array(
                [
                    [v_zero, v_one],
                    [-v_h_void + v_one_half, v_inner_c_h + v_one_half],
                    [-v_outer_c_h + v_one_half, v_one],
                    [-v_inner_c_h + v_one_half, v_h_void + v_one_half],
                ]
            )

            bottom = _np.array(
                [
                    [v_outer_c_h + v_one_half, v_zero],
                    [v_inner_c_h + v_one_half, -v_h_void + v_one_half],
                    [-v_outer_c_h + v_one_half, v_zero],
                    [-v_inner_c_h + v_one_half, -v_h_void + v_one_half],
                ]
            )

            bottom_right = _np.array(
                [
                    [v_inner_c_h + v_one_half, -v_h_void + v_one_half],
                    [v_outer_c_h + v_one_half, v_zero],
                    [v_h_void + v_one_half, -v_inner_c_h + v_one_half],
                    [v_one, -v_outer_c_h + v_one_half],
                ]
            )

        elif closure == "x_max":
            right = _np.array(
                [
                    [v_h_void + v_one_half, -v_inner_c_h + v_one_half],
                    [v_one, v_zero],
                    [v_h_void + v_one_half, v_inner_c_h + v_one_half],
                    [v_one, v_one],
                ]
            )

            right_top = _np.array(
                [
                    [v_h_void + v_one_half, v_inner_c_h + v_one_half],
                    [v_one, v_one],
                    [v_inner_c_h + v_one_half, v_h_void + v_one_half],
                    [v_outer_c_h + v_one_half, v_one],
                ]
            )

            top = _np.array(
                [
                    [v_inner_c_h + v_one_half, v_h_void + v_one_half],
                    [v_outer_c_h + v_one_half, v_one],
                    [-v_inner_c_h + v_one_half, v_h_void + v_one_half],
                    [-v_outer_c_h + v_one_half, v_one],
                ]
            )

            bottom_left = _np.array(
                [
                    [-v_h_void + v_one_half, -v_inner_c_h + v_one_half],
                    [v_zero, -v_outer_c_h + v_one_half],
                    [-v_inner_c_h + v_one_half, -v_h_void + v_one_half],
                    [-v_outer_c_h + v_one_half, v_zero],
                ]
            )

            left = _np.array(
                [
                    [v_zero, -v_outer_c_h + v_one_half],
                    [-v_h_void + v_one_half, -v_inner_c_h + v_one_half],
                    [v_zero, v_outer_c_h + v_one_half],
                    [-v_h_void + v_one_half, v_inner_c_h + v_one_half],
                ]
            )

            top_left = _np.array(
                [
                    [v_zero, v_outer_c_h + v_one_half],
                    [-v_h_void + v_one_half, v_inner_c_h + v_one_half],
                    [-v_outer_c_h + v_one_half, v_one],
                    [-v_inner_c_h + v_one_half, v_h_void + v_one_half],
                ]
            )

            bottom = _np.array(
                [
                    [v_outer_c_h + v_one_half, v_zero],
                    [v_inner_c_h + v_one_half, -v_h_void + v_one_half],
                    [-v_outer_c_h + v_one_half, v_zero],
                    [-v_inner_c_h + v_one_half, -v_h_void + v_one_half],
                ]
            )

            bottom_right = _np.array(
                [
                    [v_inner_c_h + v_one_half, -v_h_void + v_one_half],
                    [v_outer_c_h + v_one_half, v_zero],
                    [v_h_void + v_one_half, -v_inner_c_h + v_one_half],
                    [v_one, v_zero],
                ]
            )

        elif closure == "y_min":
            # set points:
            right = _np.array(
                [
                    [v_h_void + v_one_half, -v_inner_c_h + v_one_half],
                    [v_one, -v_outer_c_h + v_one_half],
                    [v_h_void + v_one_half, v_inner_c_h + v_one_half],
                    [v_one, v_outer_c_h + v_one_half],
                ]
            )

            right_top = _np.array(
                [
                    [v_h_void + v_one_half, v_inner_c_h + v_one_half],
                    [v_one, v_outer_c_h + v_one_half],
                    [v_inner_c_h + v_one_half, v_h_void + v_one_half],
                    [v_outer_c_h + v_one_half, v_one],
                ]
            )

            top = _np.array(
                [
                    [v_inner_c_h + v_one_half, v_h_void + v_one_half],
                    [v_outer_c_h + v_one_half, v_one],
                    [-v_inner_c_h + v_one_half, v_h_void + v_one_half],
                    [-v_outer_c_h + v_one_half, v_one],
                ]
            )

            bottom_left = _np.array(
                [
                    [-v_h_void + v_one_half, -v_inner_c_h + v_one_half],
                    [v_zero, -v_outer_c_h + v_one_half],
                    [-v_inner_c_h + v_one_half, -v_h_void + v_one_half],
                    [v_zero, v_zero],
                ]
            )

            left = _np.array(
                [
                    [v_zero, -v_outer_c_h + v_one_half],
                    [-v_h_void + v_one_half, -v_inner_c_h + v_one_half],
                    [v_zero, v_outer_c_h + v_one_half],
                    [-v_h_void + v_one_half, v_inner_c_h + v_one_half],
                ]
            )

            top_left = _np.array(
                [
                    [v_zero, v_outer_c_h + v_one_half],
                    [-v_h_void + v_one_half, v_inner_c_h + v_one_half],
                    [-v_outer_c_h + v_one_half, v_one],
                    [-v_inner_c_h + v_one_half, v_h_void + v_one_half],
                ]
            )

            bottom = _np.array(
                [
                    [v_one, v_zero],
                    [v_inner_c_h + v_one_half, -v_h_void + v_one_half],
                    [v_zero, v_zero],
                    [-v_inner_c_h + v_one_half, -v_h_void + v_one_half],
                ]
            )

            bottom_right = _np.array(
                [
                    [v_inner_c_h + v_one_half, -v_h_void + v_one_half],
                    [v_one, v_zero],
                    [v_h_void + v_one_half, -v_inner_c_h + v_one_half],
                    [v_one, -v_outer_c_h + v_one_half],
                ]
            )

        elif closure == "y_max":
            # set points:
            right = _np.array(
                [
                    [v_h_void + v_one_half, -v_inner_c_h + v_one_half],
                    [v_one, -v_outer_c_h + v_one_half],
                    [v_h_void + v_one_half, v_inner_c_h + v_one_half],
                    [v_one, v_outer_c_h + v_one_half],
                ]
            )

            right_top = _np.array(
                [
                    [v_h_void + v_one_half, v_inner_c_h + v_one_half],
                    [v_one, v_outer_c_h + v_one_half],
                    [v_inner_c_h + v_one_half, v_h_void + v_one_half],
                    [v_one, v_one],
                ]
            )

            top = _np.array(
                [
                    [v_inner_c_h + v_one_half, v_h_void + v_one_half],
                    [v_one, v_one],
                    [-v_inner_c_h + v_one_half, v_h_void + v_one_half],
                    [v_zero, v_one],
                ]
            )

            bottom_left = _np.array(
                [
                    [-v_h_void + v_one_half, -v_inner_c_h + v_one_half],
                    [v_zero, -v_outer_c_h + v_one_half],
                    [-v_inner_c_h + v_one_half, -v_h_void + v_one_half],
                    [-v_outer_c_h + v_one_half, v_zero],
                ]
            )

            left = _np.array(
                [
                    [v_zero, -v_outer_c_h + v_one_half],
                    [-v_h_void + v_one_half, -v_inner_c_h + v_one_half],
                    [v_zero, v_outer_c_h + v_one_half],
                    [-v_h_void + v_one_half, v_inner_c_h + v_one_half],
                ]
            )

            top_left = _np.array(
                [
                    [v_zero, v_outer_c_h + v_one_half],
                    [-v_h_void + v_one_half, v_inner_c_h + v_one_half],
                    [v_zero, v_one],
                    [-v_inner_c_h + v_one_half, v_h_void + v_one_half],
                ]
            )

            bottom = _np.array(
                [
                    [v_outer_c_h + v_one_half, v_zero],
                    [v_inner_c_h + v_one_half, -v_h_void + v_one_half],
                    [-v_outer_c_h + v_one_half, v_zero],
                    [-v_inner_c_h + v_one_half, -v_h_void + v_one_half],
                ]
            )

            bottom_right = _np.array(
                [
                    [v_inner_c_h + v_one_half, -v_h_void + v_one_half],
                    [v_outer_c_h + v_one_half, v_zero],
                    [v_h_void + v_one_half, -v_inner_c_h + v_one_half],
                    [v_one, -v_outer_c_h + v_one_half],
                ]
            )

        spline_list.append(_Bezier(degrees=[1, 1], control_points=right))

        spline_list.append(_Bezier(degrees=[1, 1], control_points=right_top))

        spline_list.append(_Bezier(degrees=[1, 1], control_points=bottom))

        spline_list.append(_Bezier(degrees=[1, 1], control_points=bottom_left))

        spline_list.append(_Bezier(degrees=[1, 1], control_points=left))

        spline_list.append(_Bezier(degrees=[1, 1], control_points=top_left))

        spline_list.append(_Bezier(degrees=[1, 1], control_points=top))

        spline_list.append(
            _Bezier(degrees=[1, 1], control_points=bottom_right)
        )

        return (spline_list, None)
